from io import BytesIO
import gzip
import os
from pathlib import Path
from typing import Union


def is_gz_bytes(chkd_bytes: bytes) -> bool:
    """
    Checks if provided bytes is a gz zipped file stream. If so returns True

    Parameters
    ----------
    chkd_bytes: bytes
        bytes to check for being gz zipped file

    Returns
    -------
    bool
        flag indicating if provided bytes is a gz compressed bytes

    """

    if not isinstance(chkd_bytes, bytes):
        return False

    with gzip.GzipFile(fileobj=BytesIO(chkd_bytes)) as f:
        try:
            f.read(1)
            return True
        except Exception as exc:
            return False


def get_unzip_gz(to_be_unzpd_bytes: bytes) -> bytes:
    """
    Unzips provided input bytes

    Parameters
    ----------
    to_be_unzpd_bytes: bytes
        bytes to be unzipped

    Returns
    -------
    bytes
        unzipped bytes

    """

    if not isinstance(to_be_unzpd_bytes, bytes):
        raise TypeError('`unzpd_bytes` param should be of bytes type')

    unzipped_buffer = BytesIO(to_be_unzpd_bytes)
    unzipped_buffer.seek(0)
    res = gzip.decompress(unzipped_buffer.read())
    return res


def check_and_get_unzipped_model(model_src: Union[str, bytes]) -> Union[str, bytes, Path]:
    """
    Checks if provided model is a compressed one and unzips it if so.

    Parameters
    ----------
    model_src: Union[str, bytes]
        model source; cna be bytes or string

    Returns
    -------
    Union[str, bytes]
        either decompressed bytes or path to model file

    """

    if isinstance(model_src, bytes):
        if is_gz_bytes(chkd_bytes=model_src):
            return get_unzip_gz(to_be_unzpd_bytes=model_src)
        else:
            return model_src
    elif isinstance(model_src, str) or isinstance(model_src, Path):
        if not os.path.isfile(model_src):
            raise ValueError(
                'This is not a proper path: {} and reading model from '
                'string is not supported'.format(model_src))

        try:
            with open(model_src, 'rb') as chkd_bytes:
                read_chkd_bytes = chkd_bytes.read()
                if is_gz_bytes(chkd_bytes=read_chkd_bytes):
                    # print('Returning unzipped file')
                    return gzip.decompress(read_chkd_bytes)
        except Exception as exc:
            print(
                'When checking file: {} the following error occurred'
                .format(model_src, exc))
        return model_src
    else:
        raise TypeError(
            'Unsupported model source type: {}'.format(type(model_src)))



import improveai.decision_model as dm


class GivensProvider:

    def givens(self, for_model, givens: dict or None = None) -> dict:
        """
        Provides givens for input `for_model`. Input `givens` are included in
        returned value

        Parameters
        ----------
        for_model: dm.DecisionModel
            instance of DecisionModel for which givens should be provided
        givens: dict or None
            givens which will be returned

        Returns
        -------
        dict or None
            givens which will be returned

        """
        assert isinstance(for_model, dm.DecisionModel)
        assert isinstance(givens, dict) or givens is None
        return givens



#!python3
#cython: language_level=3

cdef extern from "npy_no_deprecated_api.h": pass

import cython
from libc.math cimport isnan
import numpy as np
cimport numpy as np
import warnings
import xxhash


cdef object xxhash3 = xxhash.xxh3_64_intdigest
cdef set JSON_SERIALIZABLE_TYPES = {int, float, str, bool, list, tuple, dict}

import improveai.cythonized_feature_encoding.cythonized_feature_encoding_utils as cfeu
import improveai.feature_encoder as fe


encoded_variant_into_np_row = cfeu.encoded_variant_into_np_row
encode_variants_multiple_givens = cfeu.encode_variants_multiple_givens
encoded_variants_to_np = cfeu.encoded_variants_to_np


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef _is_object_json_serializable(object object_):
    """
    Checks if input value is JSON serializable

    Parameters
    ----------
    object_: object
        object to be checked

    Returns
    -------
    bool
        True if input is JSON serializable False otherwise
    """
    return type(object_) in JSON_SERIALIZABLE_TYPES or object_ is None


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef _has_top_level_string_keys(dict checked_dict):
    """
    Check if all top level keys are of a string type. This is a helper function 
    for encode() and since encode recurs in case of nested dicts only top level 
    keys need to be checked. Return False on first encountered non-string key

    Parameters
    ----------
    checked_dict: dict
        dict which top level keys will be checked

    Returns
    -------
    bool
        True if all keys are strings otherwise False

    """
    return all([isinstance(k, str) for k in checked_dict.keys()])


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef warn_about_array_encoding(object object_):
    """
    If object is an array warning will be printed (only once so the encoding speed does not suffer)

    Parameters
    ----------
    object_: object
        encoded object

    Returns
    -------
    None
        None

    """
    # check if variant is a list, tuple or array
    if not fe.WARNED_ABOUT_ARRAY_ENCODING and \
            (isinstance(object_, list) or isinstance(object_, tuple) or isinstance(object_, np.ndarray)):
        warnings.warn('Array encoding may change in the near future')
        fe.WARNED_ABOUT_ARRAY_ENCODING = True


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef encode(object object_, unsigned long long seed, double small_noise, dict features):
    """
    Encodes a JSON serializable object to  a flat key - value pair(s) structure / dict
    (sometimes a single `object_` will result in 2 features, e.g. strings, lists, etc...).
    Rules of encoding go as follows:
    
    - None, json null, {}, [], nan, are treated as missing feature_names and ignored.  NaN is technically not allowed in JSON but check anyway.
    
    - boolean true and false are encoded as 1.0 and 0.0 respectively.
    
    - strings are encoded into an additional one-hot feature.
    
    - numbers are encoded as-is with up to about 5 decimal places of precision
    
    - a small amount of noise is incorporated to avoid overfitting
    
    - feature names are 8 hexadecimal characters
    

    Parameters
    ----------
    object_: object
        a JSON serializable object to be encoded to a flat key-value structure
    seed: int
        seed for xxhash3 to generate feature name
    small_noise: float
        a shrunk noise to be added to value of encoded feature
    features: dict
        a flat dict of {<feature name>: <feature value>, ...} pairs

    Returns
    -------
    None
        None

    """

    assert _is_object_json_serializable(object_=object_)
    warn_about_array_encoding(object_=object_)

    cdef str feature_name = None
    cdef double previous_object_

    if isinstance(object_, (int, float)):  # bool is an instanceof int
        if isnan(object_):  # nan is treated as missing feature, return
            return

        feature_name = hash_to_feature_name(seed)

        previous_object_ = \
            _get_previous_value(feature_name=feature_name, into=features, small_noise=small_noise)

        features[feature_name] = sprinkle(object_ + previous_object_, small_noise)

        return

    cdef unsigned long long hashed
    cdef double previous_hashed
    cdef str hashed_feature_name
    cdef double previous_hashed_for_feature_name

    if isinstance(object_, str):
        hashed = xxhash3(object_, seed=seed)

        feature_name = hash_to_feature_name(seed)

        previous_hashed = \
            _get_previous_value(feature_name=feature_name, into=features, small_noise=small_noise)

        features[feature_name] = \
            sprinkle(<double>(<long long>(((hashed & 0xffff0000) >> 16) - 0x8000)) + previous_hashed, small_noise)

        hashed_feature_name = hash_to_feature_name(hashed)

        previous_hashed_for_feature_name = \
            _get_previous_value(feature_name=hashed_feature_name, into=features, small_noise=small_noise)

        features[hashed_feature_name] = \
            sprinkle(<double>(<long long>((hashed & 0xffff) - 0x8000)) + previous_hashed_for_feature_name, small_noise)

        return

    if isinstance(object_, dict):
        assert _has_top_level_string_keys(object_)
        for key, value in object_.items():
            encode(value, xxhash3(key, seed=seed), small_noise, features)
        return

    if isinstance(object_, (list, tuple)):
        for index, item in enumerate(object_):
            encode(item, xxhash3(index.to_bytes(8, byteorder='big'), seed=seed), small_noise, features)

        return

@cython.boundscheck(False)
@cython.wraparound(False)
cpdef double _get_previous_value(str feature_name, dict into, double small_noise):
    """
    Gets previous, 'unsprinkled' value of <feature_name> feature
    from <features> dict

    Parameters
    ----------
    feature_name: str
        name of a feature which previous value is desired
    into: dict
        dict containing current results of encoding
    small_noise: float
        small noise used to sprinkle

    Returns
    -------
    float
        'unsprinkled' value of desired feature

    """

    cdef double  previous_sprinkled_object_

    if into.get(feature_name, None) is None:
        return 0.0
    else:
        previous_sprinkled_object_ = into.get(feature_name, None)
        return reverse_sprinkle(previous_sprinkled_object_, small_noise)


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef str hash_to_feature_name(unsigned long long hash_):
    """
    Converts a hash to string which will become a feature name

    Parameters
    ----------
    hash_: int
        an integer output from xxhash3

    Returns
    -------
    str
        a string representation of a hex feature name created from int

    """
    return '%0*x' % (8, (hash_ >> 32))


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef double shrink(double noise):
    """
    Shrinks noise value by a hardcoded factor 2 ** -17

    Parameters
    ----------
    noise: float
        value within 0 - 1 range which will be 'shrunk' and added to feature value


    Returns
    -------
    float
        a shrunk noise

    """
    # return noise * pow(2, -17)
    # TODO after pypi deploy make sure this is used in the trainer
    return noise * np.float_power(2, -17)


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef double sprinkle(double x, double small_noise):
    """
    Slightly modified input valuse using `small_noise`: (x + small_noise) * (1 + small_noise)

    Parameters
    ----------
    x: float
        a number to be 'modified'
    small_noise: float
        a small number with which `x` will be modified

    Returns
    -------
    float
        x modified with `small_noise`

    """
    return (x + small_noise) * (1 + small_noise)


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef double reverse_sprinkle(double sprinkled_x, double small_noise):
    """
    Retrieves true value from sprinkled one

    Parameters
    ----------
    sprinkled_x: float
        sprinkled value
    small_noise: float
        noise used to calculate sprinkled_Xx

    Returns
    -------
    float
        true value of x which was used to calculate sprinkled_x

    """
    return sprinkled_x / (1 + small_noise) - small_noise


cdef class FeatureEncoder:
    """
    This class serves as a preprocessor which allows to
    convert event data in form of a JSON line (variants, givens and
    extra features) to the tabular format expected on input by XGBoost.

    Short summary of methods:
     - encode_givens() -> encodes givens / variant metadata to dict of
       feature name -> float pairs

     - encode_variant() -> encodes variant to dict of feature name -> float pairs

     - encode_feature_vector() -> encodes provided variant and givens;
       appends extra features to the encoded variant dict;
       converts encoded variant to numpy array using provided feature names

     - add_noise() -> sprinkles each value of the input dict with small noise

     - encode_to_np_matrix()  -> encodes collection of variants with collection
       of givens and extra features to the numpy 2D array / matrix. If desired
       uses cython backed (works much faster)

     - add_extra_features() - helper method for adding extra features to already
       encoded variants

    """

    cdef unsigned long long variant_seed
    cdef unsigned long long value_seed
    cdef unsigned long long givens_seed

    def __init__(self, long long model_seed):
        """
        Init with params

        Parameters
        ----------
        model_seed: int
            model seed to be used during feature encoding
        """
        if (model_seed < 0):
            raise TypeError(
                "xxhash3_64 takes an unsigned 64-bit integer as seed. "
                "Seed must be greater than or equal to 0.")

        # memoize commonly used seeds
        self.variant_seed = xxhash3("variant", seed=model_seed)
        self.value_seed = xxhash3("$value", seed=self.variant_seed)
        self.givens_seed = xxhash3("givens", seed=model_seed)

    cdef void _check_noise_value(self, double noise) except *:
        """
        Check whether noise value is valid; Raises if `noise` is invalid

        Parameters
        ----------
        noise: float
            checked value of no ise

        Returns
        -------
        None
            None
        """

        if noise > 1.0 or noise < 0.0:
            raise ValueError(
                'Provided `noise` is out of allowed bounds <0.0, 1.0>')

    cpdef dict encode_givens(self, dict givens, double noise=0.0, dict into=None):
        """
        Encodes provided givens of arbitrary complexity to a flat feature dict;
        Givens must be JSON encodable

        Parameters
        ----------
        givens: dict or None
            givens to be encoded
        noise: float
            value within 0 - 1 range which will be 'shrunk' and added to feature value
        into: dict
            a dict into which new features will be inserted; Can be None

        Returns
        -------
        dict
            A dict with results of givens encoding

        """

        self._check_noise_value(noise=noise)

        if givens is not None and not isinstance(givens, dict):
            raise TypeError(
                "Only dict type is supported for context encoding. {} type was "
                "provided.".format(type(givens)))

        if into is None:
            into = {}

        encode(givens, self.givens_seed, shrink(noise), into)

        return into

    cpdef dict encode_variant(self, object variant, double noise=0.0, dict into=None):
        """
        Encodes provided variant of arbitrary complexity to a flat feature dict;
        Variant must be JSON encodable

        Parameters
        ----------
        variant: object
            variant to be encoded
        noise: float
            value within 0 - 1 range which will be 'shrunk' and added to feature value
        into: dict
            a dict into which new features will be inserted; Can be None

        Returns
        -------
        dict
            A dict with results of variant encoding

        """

        self._check_noise_value(noise=noise)

        if into is None:
            into = {}

        small_noise = shrink(noise)

        if isinstance(variant, dict):
            encode(variant, self.variant_seed, small_noise, into)
        else:
            encode(variant, self.value_seed, small_noise, into)

        return into


    cpdef encode_feature_vector(
            self, object variant = None, dict givens = None,
            dict extra_features = None, list feature_names = None,
            double noise = 0.0, np.ndarray into = None):
        """
        Fully encodes provided variant and givens into a np.ndarray provided as `into` parameter.
        `into` must not be None

        Parameters
        ----------
        variant: object
            a JSON encodable object to be encoded to flat features' dict
        givens: dict
            a dict with givens to be enncoded to flat features' dict (all entries must be JSON encodable)
        extra_features: dict
            features to be added to encoded variant with givens
        feature_names: list or np.ndarray
            list of model's feature names (only overlapping features will be selected for resulting vector)
        noise: float
            value within 0 - 1 range which will be 'shrunk' and added to feature value
        into: np.ndarray
            an array into which feature values will be added / inserted

        Returns
        -------
        None
            None

        """

        if into is None:
            raise ValueError('`into` can`t be None')

        cdef dict encoded_givens = {}
        # encode givens
        encoded_givens = \
            self.encode_givens(givens=givens, noise=noise, into=dict())

        cdef dict encoded_variant = {}
        # encoded variant and givens
        encoded_variant = \
            self.encode_variant(
                variant=variant, noise=noise, into=encoded_givens)

        if extra_features:
            if not isinstance(extra_features, dict):
                raise TypeError(
                    'Provided `extra_features` is not a dict but: {}'
                    .format(extra_features))

            encoded_variant.update(extra_features)

        # n + nan = nan so you'll have to check for nan values on into
        encoded_variant_into_np_row(
            encoded_variant=encoded_variant, feature_names=feature_names,
            into=into)

    cpdef double [:, :] encode_to_np_matrix(
            self, list variants, list multiple_givens,
            list multiple_extra_features, list feature_names, double noise):
        """
        Provided list of variants and corresponding lists of givens and extra
        features encodes variants completely and converts to numpy array

        Parameters
        ----------
        variants: list
            list of variants to be encoded
        multiple_givens: list
            list of givens - givens[i] will be used to encode variants[i]
        multiple_extra_features: list
            list of extra_features - multiple_extra_features[i] will be used
            to encode variants[i]
        feature_names: list
            names of features expected by the model
        noise: float
            noise to be used when encoding data

        Returns
        -------
        np.ndarray
            2D numpy array of encoded variants

        """

        encoded_variants = encode_variants_multiple_givens(
            variants=variants, multiple_givens=multiple_givens,
            multiple_extra_features=multiple_extra_features, noise=noise,
            variants_encoder=self.encode_variant,
            givens_encoder=self.encode_givens)
        encoded_variants_array = encoded_variants_to_np(
            encoded_variants=encoded_variants, feature_names=feature_names)

        return encoded_variants_array

    cpdef void add_extra_features(self, list encoded_variants, list extra_features):
        """
        Once variants are encoded this method can be used to quickly append
        extra features

        Parameters
        ----------
        encoded_variants: list
            collection of encoded variants to be updated with extra features
        extra_features: list
            payload to be appended to encoded variants

        Returns
        -------
        None
            None

        """

        if extra_features is None:
            return

        if not isinstance(encoded_variants, list):
            raise TypeError('`encoded_variants` should be of a list type')

        if not isinstance(extra_features, list):
            raise TypeError('`extra_features` should be of a list type')

        [encoded_variant.update(single_extra_features)
         if single_extra_features is not None else encoded_variant
         for encoded_variant, single_extra_features in
         zip(encoded_variants, extra_features)]



from collections.abc import Iterable
from copy import deepcopy
import json
import numpy as np
from pathlib import Path
import pickle
import re
from traceback import print_exc
from xgboost import Booster, DMatrix
from xgboost.core import XGBoostError

from improveai.feature_encoder import FeatureEncoder
from improveai.settings import CYTHON_BACKEND_AVAILABLE
from improveai.utils.choosers_feature_encoding_tools import encoded_variants_to_np
from improveai.utils.gzip_tools import check_and_get_unzipped_model
from improveai.utils.url_tools import is_path_http_addr, get_model_bytes_from_url


if CYTHON_BACKEND_AVAILABLE:
    from improveai.cythonized_feature_encoding import cfe, cfeu
    FastFeatureEncoder = cfe.FeatureEncoder
    fast_encoded_variants_to_np = cfeu.encoded_variants_to_np
    fast_encode_variants_single_givens = cfeu.encode_variants_single_givens
else:
    FastFeatureEncoder = FeatureEncoder


class XGBChooser:
    MODEL_NAME_REGEXP = "^[a-zA-Z0-9][\w\-.]{0,63}$"
    """
    Model name regexp used to verify all model names (both user provided and cached in boosters)
    """

    @property
    def model(self) -> Booster:
        """
        xgboost's booster used by this chooser

        Returns
        -------
        Booster
            xgboost's booster used by this chooser

        """
        return self._model

    @model.setter
    def model(self, new_val: Booster):
        self._model = new_val

    @property
    def model_metadata(self) -> dict:
        """
        Improve AI model metadata dict

        Returns
        -------
        dict
            Improve AI model metadata dict

        """
        return self._model_metadata

    @model_metadata.setter
    def model_metadata(self, new_val: dict):
        self._model_metadata = new_val

    @property
    def feature_encoder(self) -> FeatureEncoder or FastFeatureEncoder:
        """
        FeatureEncoder of this chooser

        Returns
        -------
        FeatureEncoder
            FeatureEncoder of this chooser

        """
        return self._feature_encoder

    @feature_encoder.setter
    def feature_encoder(self, new_val: FeatureEncoder or FastFeatureEncoder):
        self._feature_encoder = new_val

    @property
    def model_metadata_key(self):
        """
        'User defined' metadata is stored inside Improve AI Booster. Inside a
        'user defined' metadata 'model metadata' is stored under `model_metadata_key`

        Returns
        -------
        str
            `model_metadata` key inside `user_defined_metadata`

        """
        return self._mlmodel_metadata_key

    @model_metadata_key.setter
    def model_metadata_key(self, new_val: str):
        self._mlmodel_metadata_key = new_val

    @property
    def model_seed_key(self) -> str:
        """
        `model_seed` key  in `model_metadata` dict

        Returns
        -------
        str
            `model_seed` key  in `model_metadata` dict

        """
        return self._model_seed_key

    @model_seed_key.setter
    def model_seed_key(self, new_val: str):
        self._model_seed_key = new_val

    @property
    def model_seed(self):
        """
        Model seed needed for FeatureEncoder constructor

        Returns
        -------
        int
            Model seed needed for FeatureEncoder constructor

        """
        return self._model_seed

    @model_seed.setter
    def model_seed(self, value):
        self._model_seed = value

    @property
    def model_name_key(self):
        """
        `model_name` key  in `model_metadata` dict

        Returns
        -------
        str
            `model_name` key  in `model_metadata` dict

        """
        return self._model_name_key

    @model_name_key.setter
    def model_name_key(self, value):
        self._model_name_key = value

    @property
    def model_name(self):
        """
        Model name of this Improve AI model

        Returns
        -------
        str
            Model name of this Improve AI model

        """
        return self._model_name

    @model_name.setter
    def model_name(self, value):
        assert value is not None
        assert isinstance(value, str)
        assert re.search(XGBChooser.MODEL_NAME_REGEXP, value) is not None

        self._model_name = value

    @property
    def model_feature_names_key(self) -> str:
        """
        `model_feature_names` key  in `model_metadata` dict

        Returns
        -------
        str
            `model_feature_names` key  in `model_metadata` dict

        """
        return self._model_feature_names_key

    @model_feature_names_key.setter
    def model_feature_names_key(self, new_val: str):
        self._model_feature_names_key = new_val

    @property
    def model_feature_names(self) -> list:
        """
        Feature names of this Improve AI model

        Returns
        -------
        list
            Feature names of this Improve AI model

        """
        return self._model_feature_names

    @model_feature_names.setter
    def model_feature_names(self, new_val: list):
        self._model_feature_names = new_val

    @property
    def current_noise(self):
        """
        Currently used noise value. Needed for SDK + synthetic model validation

        Returns
        -------
        float
            Currently used noise

        """
        return self._current_noise

    @current_noise.setter
    def current_noise(self, value):
        self._current_noise = value

    @property
    def imposed_noise(self):
        """
        Forced noise value. Needed for SDK + synthetic model validation

        Returns
        -------
        float
            Forced noise value

        """
        return self._imposed_noise

    @imposed_noise.setter
    def imposed_noise(self, value):
        # assert noise is valid
        assert not isinstance(value, bool)
        assert isinstance(value, float) or isinstance(value, int)
        assert 0 <= value <= 1
        self._imposed_noise = value

    @property
    def improveai_major_version_from_metadata(self) -> str or None:
        """
        Stores the Improve AI model version

        Returns
        -------
        str or None
            string version of the Improve AI model or None
        """
        return self._improveai_model_version

    @improveai_major_version_from_metadata.setter
    def improveai_major_version_from_metadata(self, value):
        self._improveai_model_version = value

    @property
    def MODEL_METADATA_KEY(self):
        """
        Key in booster.attr('user_defined_metadata') storing model metadata

        Returns
        -------
        str
            json string

        """
        return 'json'

    @property
    def MODEL_FEATURE_NAMES_KEY(self):
        """
        Key in model metadata storing feature names

        Returns
        -------
        str
            'feature_names'

        """
        return 'feature_names'

    @property
    def MODEL_SEED_KEY(self):
        """
        Key in model metadata storing model seed

        Returns
        -------
        str
            'model_seed'

        """
        return 'model_seed'

    @property
    def MODEL_NAME_KEY(self):
        """
        Key in model metadata storing model name

        Returns
        -------
        str
            'model_name'

        """
        return 'model_name'

    @property
    def IMPROVE_AI_ALLOWED_MAJOR_VERSION(self):
        """
        Latest supported major model version

        Returns
        -------
        int
            7

        """
        return 7

    @property
    def IMPROVEAI_VERSION_KEY(self):
        """
        model metadata key storing model version

        Returns
        -------
        str
            'ai.improve.version'

        """
        return 'ai.improve.version'

    def __init__(self):
        """
        Initialize chooser object
        """

        self.model = None
        self.model_metadata = None

        self.feature_encoder = None
        self.model_feature_names = np.empty(shape=(1,))

        self.model_seed = None
        self._model_name = None
        self.current_noise = None
        self._imposed_noise = None
        self._improveai_major_version_from_metadata = None

    def load_model(self, input_model_src: str, verbose: bool = False):
        """
        Loads desired model from input path.

        Parameters
        ----------
        input_model_src: str
            URL / path to desired model
        verbose: bool
            should I print debug messages
        Returns
        -------
        None
            None
        """

        try:
            if verbose:
                print('Attempting to load: {} model'.format(
                    input_model_src if len(input_model_src) < 100 else
                    str(input_model_src[:10]) + ' ... ' + str(
                        input_model_src[-10:])))

            model_src = \
                input_model_src if isinstance(input_model_src, str) or isinstance(input_model_src, Path) \
                else bytearray(input_model_src)

            self.model = Booster()
            self.model.load_model(model_src)
            if verbose:
                print('Model: {} successfully loaded'.format(
                    input_model_src if len(input_model_src) < 100 else
                    str(input_model_src[:10]) + ' ... ' + str(
                        input_model_src[-10:])))
        except XGBoostError as xgbe:
            if verbose:
                print('Attempting to read via pickle interface')
            with open(input_model_src, 'rb') as xgbl:
                self.model = pickle.load(xgbl)
            print_exc()
        except Exception as exc:
            if verbose:
                print(
                    'When attempting to load the model: {} the following error '
                    'occured: {}'.format(input_model_src, exc))
            print_exc()

        model_metadata = self._get_model_metadata()
        self.model_seed = self._get_model_seed(model_metadata=model_metadata)
        self.model_name = self._get_model_name(model_metadata=model_metadata)
        self.model_feature_names = \
            self._get_model_feature_names(model_metadata=model_metadata)
        self.improveai_major_version_from_metadata = \
            self._get_improveai_major_version(model_metadata=model_metadata)

        if CYTHON_BACKEND_AVAILABLE:
            self.feature_encoder = FastFeatureEncoder(model_seed=self.model_seed)
        else:
            self.feature_encoder = FeatureEncoder(model_seed=self.model_seed)

    def _get_improveai_major_version(self, model_metadata: dict) -> str or None:
        """
        Extract Improve AI version from model metadata and return it if it is valid / allowed

        Parameters
        ----------
        model_metadata: dict
            a dictionary containing model metadata

        Returns
        -------
        str or None
            major Improve AI version extracted from loaded improve model
        """
        improveai_major_version = None
        if self.IMPROVEAI_VERSION_KEY in model_metadata.keys():
            improveai_version = model_metadata[self.IMPROVEAI_VERSION_KEY]

            print('### improveai_version ###')
            print(improveai_version)

            assert improveai_version is not None and isinstance(improveai_version, str)
            # major version is the first chunk of version string
            improveai_major_version = int(improveai_version.split('.')[0])
            assert improveai_major_version == self.IMPROVE_AI_ALLOWED_MAJOR_VERSION
        return improveai_major_version

    def _get_model_metadata(self) -> dict:
        """
        gets 'model metadata' from 'user defined metadata' of Improve AI model

        Returns
        -------
        dict
            dict with model metadata
        """

        assert 'user_defined_metadata' in self.model.attributes().keys()
        user_defined_metadata_str = self.model.attr('user_defined_metadata')
        user_defined_metadata = json.loads(user_defined_metadata_str)
        assert self.MODEL_METADATA_KEY in user_defined_metadata.keys()

        return user_defined_metadata[self.MODEL_METADATA_KEY]

    def score(self, variants: list or tuple or np.ndarray, givens: dict or None, **kwargs) -> np.ndarray:
        """
        Scores all provided variants

        Parameters
        ----------
        variants: list or tuple or np.ndarray
            list of variants to scores
        givens: dict or None
            context dict needed for encoding
        kwargs: dict
            kwargs

        Returns
        -------
        np.ndarray
            2D numpy array which contains (variant, score) pair in each row
        """

        encoded_variants = \
            self.encode_variants_single_givens(variants=variants, givens=givens)

        encoded_variants_to_np_method = \
            fast_encoded_variants_to_np if CYTHON_BACKEND_AVAILABLE \
            else encoded_variants_to_np

        missings_filled_v = \
            encoded_variants_to_np_method(
                encoded_variants=encoded_variants,
                feature_names=self.model_feature_names)

        if CYTHON_BACKEND_AVAILABLE:
            missings_filled_v = np.asarray(missings_filled_v)

        scores = \
            self.model.predict(
                DMatrix(
                    missings_filled_v, feature_names=self.model_feature_names))\
            .astype('float64')

        return scores

    def fill_missing_features(self, encoded_variants):
        """
        Fills missing features in encoded variants and packs them into 2D numpy array

        Parameters
        ----------
        encoded_variants: list
            a list of encoded variants (dicts)

        Returns
        -------
        np.ndarray
            2D numpy array with all features for xgb model

        """
        encoded_variants_to_np_method = \
            fast_encoded_variants_to_np if CYTHON_BACKEND_AVAILABLE else encoded_variants_to_np

        features_matrix = encoded_variants_to_np_method(
            encoded_variants=encoded_variants, feature_names=self.model_feature_names)

        if CYTHON_BACKEND_AVAILABLE:
            features_matrix = np.asarray(features_matrix)

        return features_matrix

    def calculate_predictions(self, features_matrix: np.ndarray):
        """
        Calculates predictions on provided matrix with loaded model

        Parameters
        ----------
        features_matrix: np.ndarray
            array to be a source for DMatrix

        Returns
        -------

        """
        # make sure input is a numpy array
        assert isinstance(features_matrix, np.ndarray)
        # make sure input for predictions is not empty
        assert features_matrix.size > 0
        # make sure it is 2D array
        assert len(features_matrix.shape) == 2
        # make sure all features are present
        assert len(self.model_feature_names) == features_matrix.shape[1]
        scores = \
            self.model.predict(
                DMatrix(features_matrix, feature_names=self.model_feature_names)) \
            .astype('float64')
        return scores

    def encode_variants_single_givens(
            self, variants: list or tuple or np.ndarray, givens: dict or None) -> Iterable:
        """
        Implemented as a XGBChooser helper method
        Cythonized loop over provided variants and a single givens dict.
        Returns array of encoded dicts.

        Parameters
        ----------
        variants: list or tuple or np.ndarray
            collection of input variants to be encoded
        givens: dict or None
            context to be encoded with variants

        Returns
        -------
        np.ndarray
            array of encoded dicts
        """

        if not (isinstance(givens, dict) or givens is None or givens is {}):
            raise TypeError(
                'Unsupported givens` type: {}'.format(type(givens)))

        if self.imposed_noise is None:
            noise = np.random.rand()
        else:
            noise = self.imposed_noise

        if CYTHON_BACKEND_AVAILABLE:
            if isinstance(variants, list):
                used_variants = variants
            elif isinstance(variants, tuple) or isinstance(variants, np.ndarray):
                used_variants = list(variants)
            else:
                raise TypeError(
                    'Variants are of a wrong type: {}'.format(type(variants)))

            return fast_encode_variants_single_givens(
                variants=used_variants, givens=givens, noise=noise,
                variants_encoder=self.feature_encoder.encode_variant,
                givens_encoder=self.feature_encoder.encode_givens)
        else:
            encoded_variants = np.empty(len(variants), dtype=object)

            encoded_givens = self.feature_encoder.encode_givens(givens=givens, noise=noise)

            encoded_variants[:] = [
                self.feature_encoder.encode_variant(
                    variant=variant, noise=noise, into=deepcopy(encoded_givens))
                for variant in variants]

            return encoded_variants

    @staticmethod
    def get_model_src(model_src: str or bytes) -> str or bytes:
        """
        Gets model src from provided input path, url or bytes

        Parameters
        ----------
        model_src: str or bytes
            pth to model, url or bytes

        Returns
        -------
        str or bytes
            path or downloaded model

        """
        raw_model_src = model_src
        if not isinstance(model_src, Path) and is_path_http_addr(pth_to_model=model_src):
            raw_model_src = get_model_bytes_from_url(model_url=model_src)

        unzipped_model_src = check_and_get_unzipped_model(model_src=raw_model_src)
        return unzipped_model_src

    def _get_model_feature_names(self, model_metadata: dict) -> list:
        """
        Gets model feature names from model metadata

        Parameters
        ----------
        model_metadata: dict
            a dict containing model metadata

        Returns
        -------
        list
            list of feature names

        """

        if not model_metadata:
            raise ValueError('Model metadata empty or None!')

        feature_names = model_metadata.get(self.MODEL_FEATURE_NAMES_KEY, None)

        if not feature_names:
            raise ValueError('Feature names not in model metadata!')

        return feature_names

    def _get_model_seed(self, model_metadata: dict) -> int:
        """
        Gets model seed from model metadata

        Parameters
        ----------
        model_metadata: dict
            a dict containing model metadata


        Returns
        -------
        int
            model seed

        """
        if not model_metadata:
            raise ValueError('Model metadata empty or None!')

        model_seed = model_metadata.get(self.MODEL_SEED_KEY, None)

        if not model_seed:
            raise ValueError('Feature names not in model metadata!')

        return model_seed

    def _get_model_name(self, model_metadata: dict) -> str:
        """
        Gets model name from model metadata

        Parameters
        ----------
        model_metadata: dict
            a dict containing model metadata

        Returns
        -------
        str
            Improve AI model name

        """
        if not model_metadata:
            raise ValueError('Model metadata empty or None!')

        model_name = model_metadata.get(self.MODEL_NAME_KEY, None)

        if not model_name:
            raise ValueError('Feature names not in model metadata!')

        return model_name

    def sort(
            self, variants_w_scores: np.ndarray,
            scores_col_idx: int = 1, class_cols_idx: int = 2) -> np.ndarray:
        """
        Performs sorting of provided variants with scores array

        Parameters
        ----------
        variants_w_scores: np.ndarray
            array with variant, scores rows
        scores_col_idx: int
            the index of column with scores
        class_cols_idx: int
            index of the class label in a single row

        Returns
        -------
        np.ndarray
            2D sorted array of rows (variant, score)

        """

        desc_scores_sorting_col = -1 * variants_w_scores[:, scores_col_idx]
        class_sorting_col = variants_w_scores[:, class_cols_idx]

        ind = np.lexsort((desc_scores_sorting_col, class_sorting_col))

        srtd_variants_w_scores = \
            variants_w_scores[ind]

        return srtd_variants_w_scores



import numpy as np
import os
import pyximport

from improveai.settings import CYTHON_BACKEND_AVAILABLE

if CYTHON_BACKEND_AVAILABLE:

    CYTHON_ENCODING_DIR_PATH = os.sep.join(__file__.split(os.sep)[:-1])
    CYTHON_MODULE_DIR = 'cythonized_feature_encoding'
    IMPROVE_DIR = 'improveai'

    try:
        pyximport.install(setup_args={'include_dirs': [np.get_include(), CYTHON_ENCODING_DIR_PATH]})
        # this might raise a ValueError if numpy version differs from the one which was used to build improveai
        import improveai.cythonized_feature_encoding.cythonized_feature_encoding_utils as cfeu
        import improveai.cythonized_feature_encoding.cythonized_feature_encoder as cfe
    except ValueError as verr:
        # if numpy current numpy version is different than the one which was used to build improveai
        # numpy headers might not match in compiled files
        # to address this rebuilding of *.c files is needed
        files_to_delete = \
            [file_name for file_name in os.listdir(CYTHON_ENCODING_DIR_PATH)
             if file_name.endswith('.c') or file_name.endswith('.so')]
        [os.remove(os.sep.join([CYTHON_ENCODING_DIR_PATH, cython_compiled_file]))
         for cython_compiled_file in files_to_delete]
        pyximport.install(setup_args={'include_dirs': [np.get_include(), CYTHON_ENCODING_DIR_PATH]})
        import improveai.cythonized_feature_encoding.cythonized_feature_encoding_utils as cfeu
        import improveai.cythonized_feature_encoding.cythonized_feature_encoder as cfe





from pathlib import Path
import re
import warnings

import numpy as np

from improveai.chooser import XGBChooser
import improveai.decision as d
import improveai.decision_context as dc
import improveai.decision_tracker as dt
import improveai.givens_provider as gp
from improveai.settings import DEBUG
from improveai.utils.general_purpose_tools import check_variants, \
    get_variants_from_args, is_valid_ksuid, is_valid_variants_type, ALLOWED_VARIANT_COLLECTION_TYPES


class DecisionModel:

    SUPPORTED_CALLS = ['score', 'rank', 'get', 'which_from', 'optimize']
    """
    Method calls supported by Improve model CLI
    """

    MODEL_NAME_REGEXP = XGBChooser.MODEL_NAME_REGEXP
    """
    Model name regexp used to verify all model names (both user provided and cached in boosters)
    """

    @property
    def model_name(self) -> str or None:
        """
        Name of DecisionModel. None value is allowed but otherwise value must be
        a string and pass `MODEL_NAME_REGEXP` check

        Returns
        -------
        str of None
            `model_name` of this DecisionModel

        """

        return self.__model_name

    @property
    def chooser(self) -> XGBChooser:
        """
        Chooser used by current DecisionModel. Currently only XGBChooser is supported.
        If not set will return None

        Returns
        -------
        XGBChooser or None
            DecisionModel's chooser

        """
        return self._chooser

    @chooser.setter
    def chooser(self, value: XGBChooser or None):
        self._chooser = value

    @property
    def givens_provider(self) -> gp.GivensProvider:
        """
        GivensProvider used by this DecisionModel

        Returns
        -------
        GivensProvider
            GivensProvider used by this DecisionModel

        """
        return self._givens_provider

    @givens_provider.setter
    def givens_provider(self, value):
        self._givens_provider = value

    @property
    def track_url(self) -> str or None:
        """
        Track endpoint URL used by this model. Also updates `track_url` of a DecisionTracker

        Returns
        -------
        str or None
            track endpoint URL string

        """
        return self._track_url

    @track_url.setter
    def track_url(self, value: str or None):
        self._track_url = value

        # cases:
        # 1. __tracker is set and is not None
        # 2. __tracker is set and is None
        # 3. __tracker is not set

        # additionally self.track_api_key:
        # 1. is set
        # 2. is not set

        # Allow None value in constructor
        if hasattr(self, '_DecisionModel__tracker') and getattr(self, '_DecisionModel__tracker') is not None:
            self.__tracker.track_url = value
        elif (not hasattr(self, '_DecisionModel__tracker') or self.__tracker is None) and value is not None:
            self.__tracker = dt.DecisionTracker(
                track_url=value,
                track_api_key=self.track_api_key if hasattr(self, '_DecisionModel__track_api_key') else None)

    @property
    def track_api_key(self):
        """
        track API key to be sent as the x-api-key HTTP header on track requests. Omit header if null

        Returns
        -------
        str
            current track API key

        """
        return self.__track_api_key

    @track_api_key.setter
    def track_api_key(self, value):
        """
        sets track API key to be sent as the x-api-key HTTP header on track requests. Omit header if null

        Parameters
        ----------
        value: str
            track API key for this instance

        """
        # set API key to decision model
        self.__track_api_key = value
        # update API key on the tracker
        if hasattr(self, '_DecisionModel__tracker') and getattr(self, '_DecisionModel__tracker') is not None:
            self.__tracker.api_key = value

    @property
    def tracker(self) -> dt.DecisionTracker:
        """
        DecisionTracker of this DecisionModel

        Returns
        -------
        DecisionTracker or None
            `tracker` of current DecisionModel or None if `tracker` was not set

        """
        return self.__tracker

    @property
    def last_decision_id(self):
        """
        DecisionModel persists last tracked decision ID in `last_decision_id` property

        Returns
        -------
        str
            Ksuid as string

        """
        return self._last_decision_id

    @last_decision_id.setter
    def last_decision_id(self, value):
        is_valid_ksuid(value)
        self._last_decision_id = value

    @property
    def TIEBREAKER_MULTIPLIER(self) -> float:
        """
        Small value randomized and added to model's scores

        Returns
        -------
        float
            Small value randomized and added to model's scores

        """
        return 2**-23

    def __init__(self, model_name: str, track_url: str = None, track_api_key: str = None):
        """
        Init with params

        Parameters
        ----------
        model_name: str
            model name for this DecisionModel; can be None
        track_url: str
            a valid URL leading to improve gym track endpoint
        track_api_key: str
            track endpoint API key (if available); can be None
        """

        self.__check_model_name(model_name=model_name)
        self.__model_name = model_name

        self.track_url = track_url
        self.track_api_key = track_api_key

        if not hasattr(self, '_DecisionModel__tracker'):
            self.__tracker = None

        self.chooser = None
        self.givens_provider = gp.GivensProvider()

        # init `self._last_decision_id` to None
        self._last_decision_id = None

    def __check_model_name(self, model_name: str):
        """
        Check if provided model name is a valid model name

        Parameters
        ----------
        model_name: str
            checked model name

        """
        if model_name is not None:
            assert isinstance(model_name, str), \
                f'Invalid model name type - expected string got {type(model_name)}'
            assert re.search(DecisionModel.MODEL_NAME_REGEXP, model_name) is not None, \
                f'Valid model name must pass regexpr: {DecisionModel.MODEL_NAME_REGEXP}'

    def load(self, model_url: str or Path):
        """
        Synchronously loads XGBoost model from provided path, creates instance
        of DecisionModel and returns it

        Parameters
        ----------
        model_url: str
            path to desired model (FS path or url)

        Returns
        -------
        DecisionModel
            new instance of decision model

        """
        self.chooser = self._get_chooser(model_url=model_url)
        self._resolve_model_name()

        return self

    def _is_loaded(self):
        """
        Checks if a valid decision model was loaded

        Returns
        -------
        bool
            was model loaded or not

        """
        return self.chooser is not None and isinstance(self.chooser, XGBChooser)

    def _resolve_model_name(self):
        """
        If model name is not set loaded model name is used, otherwise model name
        from constructor `wins`

        Returns
        -------

        """

        if self.model_name is None:
            self.__check_model_name(model_name=self.chooser.model_name)
            self.__model_name = self.chooser.model_name
        else:
            self.chooser.model_name = self.model_name

    def _get_chooser(self, model_url: str) -> XGBChooser:
        """
        Synchronously loads XGBoost model from provided path, creates instance
        of Chooser and returns it

        Parameters
        ----------
        model_url: str
            path to desired model (FS path or url)

        Returns
        -------
        XGBChooser
            loaded chooser

        """

        try:
            model_src = XGBChooser.get_model_src(model_src=model_url)
        except Exception as exc:
            raise exc

        chooser = XGBChooser()
        chooser.load_model(input_model_src=model_src)
        return chooser

    def score(self, variants: list or np.ndarray) -> np.ndarray:
        """
        Scores provided variants with available givens

        Parameters
        ----------
        variants: list or tuple or np.ndarray
            array of variants to choose from

        Returns
        -------
        np.ndarray
            array of float64 scores

        """

        # use GivensProvider for givens (this will result in empty givens for now)
        # return equivalent of double scores
        return self._score(variants=variants, givens=self.givens_provider.givens(for_model=self))

    def _score(self, variants: list or np.ndarray, givens: dict or None) -> np.ndarray:
        """
        Call predict and calculate scores for provided variants

        Parameters
        ----------
        variants: list or np.ndarray
            collection of variants to be scored
        givens: dict or None
            givens for variants

        Returns
        -------
        np.ndarray
            scores

        """

        check_variants(variants=variants)
        # log givens for DEBUG == True
        if DEBUG is True:
            print(f'[DEBUG] givens: {givens}')
        if self.chooser is None:
            return DecisionModel._generate_descending_gaussians(count=len(variants))

        # encode variants with single givens
        encoded_variants = \
            self.chooser.encode_variants_single_givens(variants=variants, givens=givens)
        # fill missing features and create numpy matrix for xgboost to predict on
        features_matrix = self.chooser.fill_missing_features(encoded_variants=encoded_variants)

        try:
            # calculate predictions
            scores = self.chooser.calculate_predictions(features_matrix=features_matrix) + \
                     np.array(np.random.rand(len(variants)), dtype='float64') * \
                     self.TIEBREAKER_MULTIPLIER
        except Exception as exc:
            warnings.warn(
                'Error when calculating predictions: {}. Returning Gaussian scores'
                .format(exc))
            scores = DecisionModel._generate_descending_gaussians(count=len(variants))

        return scores.astype(np.float64)

    def _rank(self, variants: list or np.ndarray, scores: list or np.ndarray) -> np.ndarray:
        """
        Helper method to rank variants. Returns a numpy array with variants ranked from best to worst

        Parameters
        ----------
        variants: np.ndarray
            collection of variants to be ranked
        scores: list or np.ndarray
            scores for provided variants

        Returns
        -------
        np.ndarray
            sorted variants

        """

        # assure provided variants are not empty
        assert len(variants) > 0
        # make sure score have the same length that ranked variants
        assert len(variants) == len(scores)

        # convert variants to numpy array for faster sorting
        variants_np = variants if isinstance(variants, np.ndarray) else np.array(variants)
        # return descending sorted variants
        return variants_np[np.argsort(scores)][::-1].tolist()

    def rank(self, variants: list or np.ndarray) -> list:
        """
        Ranks provided variants from best to worst and creates decision from them

        Parameters
        ----------
        variants: list or np.ndarray
            variants to be ranked

        Returns
        -------
        list or np.ndarray
            best to worst ranked variants
        """
        return self.decide(variants=variants).ranked

    def given(self, givens: dict) -> dc.DecisionContext:
        """
        Wrapper for chaining.

        Parameters
        ----------
        givens: dict
            givens to be set

        Returns
        -------
        DecisionContext
            decision context for a given model

        """

        # assert givens is not None
        return dc.DecisionContext(decision_model=self, givens=givens)

    def which(self, *variants) -> tuple:
        """
        A short hand version of chooseFrom that returns the chosen result directly,
        automatically tracking the decision in the process. This would be a varargs
        version of chooseFrom that would also take an array as the only argument.
        If the only argument was not an array it would throw an exception.
        Since the Decision object is not returned, rewards would only be tracked
        through DecisionModel.addReward

        Parameters
        ----------
        variants: list or tuple or np.ndarray
            array of variants passed as positional parameters

        Returns
        -------
        object, str
            tuple with (<best variant>, <decision id>)

        """

        return self.which_from(variants=get_variants_from_args(variants))

    def which_from(self, variants: list or np.ndarray) -> tuple:
        """
        Makes a decision for provided variants and DecisionContext givens.

        Parameters
        ----------
        variants: list or np.ndarray
            variants from which best will be chosen

        Returns
        -------
        object, str
            a tuple of (<best variant>, <decision id>)

        """

        # create a Decision object
        decision = self.decide(variants)
        # track if possible -> tracker and track_url exist
        if self.track_url is not None and self.tracker is not None:
            decision.track()
        # return best variant and decision ID to allow rewarding
        return decision.best, decision.id_

    def add_reward(self, reward: float, decision_id: str):
        """
        Adds provided reward for a  given decision id.

        Parameters
        ----------
        reward: float or int
            reward to be assigned to a given decision
        decision_id: str
            ksuid of rewarded decision

        Returns
        -------

        """
        assert self.model_name is not None
        assert re.search(XGBChooser.MODEL_NAME_REGEXP, self.model_name) is not None

        # make sure provided decision id is valid
        assert decision_id is not None and is_valid_ksuid(decision_id)

        if self.__tracker is not None:
            return self.__tracker.add_reward(
                reward=reward, model_name=self.model_name, decision_id=decision_id)
        else:
            if self.__tracker is None:
                warnings.warn(
                    '`tracker` is not set (`tracker`is None) - reward not added')
            if self.track_url is None:
                warnings.warn('`track_url` is None - reward not added')

    def decide(self, variants: list or np.ndarray, scores: list or np.ndarray = None,
               ordered: bool = False):
        """
        Creates decision but does not track it by default.
        If not scores are provided and input variants are not ranked performs
        scoring and ranking. If scores are provided but variants are not ordered
        ranks variants before making a decision

        Parameters
        ----------
        variants: list or np.ndarray
            variants used to make decision
        scores: list or np.ndarray
            scores for variants
        ordered: bool
            are variants ordered from best to worst

        Returns
        -------
        d.Decision
            decision object with provided input

        """

        return self.given(givens=self.givens_provider.givens(for_model=self))\
            .decide(variants=variants, scores=scores, ordered=ordered)

    def optimize(self, variant_map: dict):
        """
        Get the best configuration for a given variants map

        Parameters
        ----------
        variant_map: dict
            mapping variants name -> variants list

        Returns
        -------
        object, str
            best variant and a decision ID

        """
        return self.which_from(variants=DecisionModel.full_factorial_variants(variant_map=variant_map))

    def choose_from(self, variants: list or np.ndarray, scores: list or np.ndarray):
        # TODO method deprecated - will be removed in v8 upgrade
        """
        Wrapper for chaining

        Parameters
        ----------
        variants: np.ndarray or list or tuple
            variants to be set
        scores: list or np.ndarray
            list of already calculated scores

        Returns
        -------
        Decision
            decision object with provided with already scored variants and best selected
        """
        return self.decide(variants=variants, scores=scores)

    def choose_first(self, variants: list or np.ndarray):
        # TODO method deprecated - will be removed in v8 upgrade
        """
        Chooses first from provided variants using gaussian scores (_generate_descending_gaussians())

        Parameters
        ----------
        variants: list or tuple or np.ndarray
            collection of variants passed as positional parameters

        Returns
        -------
        Decision
            A decision with first variants as the best one and gaussian scores
        """

        # make decision and do not track it
        return self.decide(variants=variants, ordered=True)

    def first(self, *variants):
        # TODO method deprecated - will be removed in v8 upgrade
        """
        Makes decision using first variant as best and tracks it.
        Accepts variants as pythonic args

        Parameters
        ----------
        *variants: list or tuple or np.ndarray
            collection of variants of which first will be chosen

        Returns
        -------
        object, str
            tuple with (<first variant>, <decision id>)
        """
        decision = self.decide(variants=get_variants_from_args(variants), ordered=True)
        return decision.get(), decision.id_

    def choose_random(self, variants: list or np.ndarray):
        # TODO method deprecated - will be removed in v8 upgrade
        """
        Shuffles variants to return Decision with gaussian scores and random best variant

        Parameters
        ----------
        *variants: list or tuple or np.ndarray
            collection of variants of which random will be chosen

        Returns
        -------
        Decision
            Decision with randomly chosen best variant
        """
        check_variants(variants)
        return self.decide(variants, scores=np.random.normal(size=len(variants)))

    def random(self, *variants):
        # TODO method deprecated - will be removed in v8 upgrade
        """
        Makes decision using randomly selected variant as best and tracks it.
        Accepts variants as pythonic args

        Parameters
        ----------
        variants: list or np.ndarray
            collection of variants of which first will be chosen

        Returns
        -------
        object, str
            tuple with (<random variant>, <decision id>)
        """
        # TODO test if it raises the same way choose_random raises
        check_variants(variants)
        unpacked_variants = get_variants_from_args(variants)
        decision = self.decide(
            variants=unpacked_variants,
            scores=np.random.normal(size=len(unpacked_variants)))
        return decision.get(), decision.id_

    def choose_multivariate(self, variant_map: dict) -> dict:
        # TODO method deprecated - will be removed in v8 upgrade
        """
        Chooses the best configuration from full factorial of input variants map

        Parameters
        ----------
        variant_map: dict
            map of key -> variants set

        Returns
        -------
        dict
            combination of the best variants in a dict

        """
        return self.given(givens=self.givens_provider.givens(for_model=self))\
            .choose_multivariate(variant_map=variant_map)

    def _track(self, variant: object, runners_up: list, sample: object, sample_pool_size: int):
        """
        Tracks provided variant with runners up and sample

        Parameters
        ----------
        variant: object
            variant to be tracked
        runners_up: list
            runners_up to be tracked
        sample: object
            a sample for an input variant
        sample_pool_size: int
            number of variants from which sample was drawn

        Returns
        -------
        str
            decision ID
        """

        return self.given(givens=self.givens_provider.givens(for_model=self)) \
            ._track(variant=variant, runners_up=runners_up, sample=sample, sample_pool_size=sample_pool_size)

    @staticmethod
    def _generate_descending_gaussians(count: int) -> np.ndarray:
        """
        Generates random floats and sorts in a descending fashion

        Parameters
        ----------
        count: int
            number of floats to generate

        Returns
        -------
        np.ndarray
            array of sorted floats

        """

        random_scores = np.random.normal(size=count)
        random_scores[::-1].sort()
        return random_scores

    @staticmethod
    def full_factorial_variants(variant_map: dict):
        """
        Creates full factorial from input variants map, i.e. for variants_map
        {
        'variants_0': ['01', '02'],
        'variants_1': ['11', '12', '13']}

        it creates a list of dicts:
        [
        {'variants_0': '01', 'variants_1': '11'},
        {'variants_0': '02', 'variants_1': '11'},
        {'variants_0': '01', 'variants_1': '12'},
        {'variants_0': '02', 'variants_1': '22'},
        {'variants_0': '01', 'variants_1': '13'},
        {'variants_0': '02', 'variants_1': '23'},
        ]

        Parameters
        ----------
        variant_map: dict
            map of key -> variants set

        Returns
        -------
        list
            list of 'fully factorial variants' dicts

        """
        # make sure variant map is a dict
        assert isinstance(variant_map, dict)
        # make sure that variant map is not empty
        assert len(variant_map) > 0
        # make sure that each key stores not None value (variants == None is not allowed)
        assert all(vs is not None for vs in variant_map.values())
        # make sure that all collections in variant_map are not empty
        assert all(len(vs) > 0 if type(vs) in ALLOWED_VARIANT_COLLECTION_TYPES else True for vs in variant_map.values())
        # check if all entries in variants_map are lists or tuples or np.arrays
        variants_map_fixed = {
            variants_key: variants if is_valid_variants_type(variants) else [variants]
            for variants_key, variants in variant_map.items()}
        # make sure all collections of variants are of a correct type
        [check_variants(vs) for vs in variants_map_fixed.values()]
        # 0. variant_map = {'variants_0': ['00', '01'], 'variants_1': ['11', '12', '13']}
        #
        # 1. meshgrid treats each sequence as a different 'variable' and prepares all vs all permutations
        #    meshgrid returns list of n-dimensional arrays where N is initial number of 'dimensions' (keys)
        #    in variant map (len(variant_map.keys()))
        #    variants_0 'dimension' meshgrid output ((3, 2) shape):
        #    [['00' '01']
        #     ['00' '01']
        #     ['00' '01']]
        #
        #    variants_1 'dimension' meshgrid output ((3, 2) shape):
        #    [['11' '11']
        #     ['12' '12']
        #     ['13' '13']]
        # 2. each element returned by meshgrid must be flattened ((1,) shape) and will represent a 'single
        #    dimension':
        #
        #    flattened variants_0 meshgrid result: ['00', '01', '00', '01', '00', '01']
        #    flattened variants_1 meshgrid result: ['11', '11', '12', '12', '13', '13']
        #
        #    Since each element was flattened calling np.vstack() will stack them 'vertically'
        #    (as if they were rows):
        #    [['00', '01', '00', '01', '00', '01'],
        #     ['11', '11', '12', '12', '13', '13']]
        #
        # 3. transposing np.vstack() result yields full factorial of values of variants map ->
        #    last step is to convert 'records' to dicts using variant_map keys

        all_variants = [np.array(vs, dtype='O') for vs in variants_map_fixed.values()]
        return [dict(zip(variants_map_fixed.keys(), variants)) for variants
                in np.vstack([v.flat for v in np.meshgrid(*all_variants)]).T]


def load_model(model_url: str, track_url: str = None):
    """
    Creates DecisionModel for provided model URL and track URL

    Parameters
    ----------
    model_url: str
        url to improve.ai model
    track_url: str
        url to improve.ai track endpoint

    Returns
    -------
    DecisionModel
        DecisionModel with Improve AI model and desired track_url

    """
    return DecisionModel(model_name=None, track_url=track_url).load(model_url=model_url)



import numpy as np

import improveai.decision_model as dm
from improveai.utils.general_purpose_tools import check_variants, is_valid_ksuid


class Decision:

    @property
    def givens(self) -> dict or None:
        """
        Givens used to make Decision; Can be None and {}

        Returns
        -------
        dict or None
            Givens used to make Decision

        """
        return self.__givens

    @property
    def decision_model(self):
        """
        DecisionModel used to make Decision

        Returns
        -------
        DecisionModel
            DecisionModel used to make Decision

        """
        return self.__decision_model

    @property
    def ranked(self) -> list or tuple or np.ndarray:
        """
        Variants ranked from best to worst (using scores)

        Returns
        -------
        list or tuple or np.ndarray
            Variants ranked from best to worst (using scores)

        """
        return self.__ranked

    # TODO add unittests for this
    @property
    def best(self) -> object:
        """
        Best variant for this Decision object -> first of ranked provided to constructor

        Returns
        -------
        object
            best variant for this Decision

        """
        return self.__best

    @property
    def id_(self):
        """
        ID (ksuid) if this decision

        Returns
        -------
        str
            ID of this Decision

        """
        return self.__id_

    def __init__(self, decision_model: object, ranked: list, givens: dict or None):
        """
        Init with params

        Parameters
        ----------
        decision_model: DecisionModel
            DecisionModel for this Decision
        """

        assert isinstance(decision_model, dm.DecisionModel)
        self.__decision_model = decision_model

        assert isinstance(givens, dict) or givens is None
        self.__givens = givens

        check_variants(ranked)
        self.__ranked = ranked
        self.__best = ranked[0]
        self.__id_ = None

    def add_reward(self, reward: float or int):
        """
        Adds provided reward to this Decision. Reward must be a number, must not be None, nan or +- inf

        Parameters
        ----------
        reward: float
            reward to be added

        Returns
        -------
        str
            reward request message ID

        """

        assert is_valid_ksuid(self.id_)
        assert self.decision_model.model_name is not None
        assert reward is not None
        assert isinstance(reward, float) or isinstance(reward, int)
        assert not np.isnan(reward)
        assert not np.isinf(reward)

        self.decision_model.tracker.add_reward(
            reward=reward, model_name=self.decision_model.model_name, decision_id=self.id_)

    def get(self):
        """
        return best variant for this decision
        tracks this decision

        Returns
        -------
        object
            best variant

        """

        if self.id_ is None:
            self.track()
        return self.best

    def track(self):
        """
        track this decision

        Returns
        -------
        str
            decision ID (ksuid) obtained during tracking

        """
        # make sure variants are not None
        assert self.ranked is not None
        # make sure that self.__id_ is set for the first time
        assert self.id_ is None, \
            f'This decision has already an ID set: {self.id_} which means it has already been tracked'
        # TODO make sure tracker is ready
        assert self.decision_model.tracker is not None
        # track() message ID for current decision -> decision ID
        self.__id_ = self.decision_model.tracker.track(
            ranked_variants=self.ranked, givens=self.givens, model_name=self.decision_model.model_name)
        # if self.id_ is not None at this point it means that track() was called successfully
        assert self.id_ is not None
        # cache most recent tracked decision ID to a decision model
        self.decision_model.last_decision_id = self.id_

        return self.id_

    def peek(self) -> object:
        """
        Returns best variant without tracking it

        Returns
        -------
        object
            best variant for this Decision

        """
        return self.best



import time
from argparse import ArgumentParser
import json
import numpy as np
import os

from improveai import DecisionModel, DecisionContext
from improveai.utils.general_purpose_tools import read_jsonstring_from_file


def mockup_and_execute(executed_function, function_kwargs, mockup_track_endpoint, track_url):
    if mockup_track_endpoint:
        with rqm.Mocker() as m:
            m.post(track_url, text='success')
            result = executed_function(**function_kwargs)
            time.sleep(0.175)
            return result
    else:
        return executed_function(**function_kwargs)


if __name__ == '__main__':

    ap = ArgumentParser()
    ap.add_argument(
        '--operation', default='score',
        help='Which action of 4 available should be performed: {}'.format(DecisionModel.SUPPORTED_CALLS))
    ap.add_argument('--model_url', help='Path to model file', required=True)
    ap.add_argument(
        '--variants', help='Path to JSONfile with variants or a JSON itself',
        default='[]')

    ap.add_argument(
        '--givens',
        help='Path to JSONfile with givens or JSON with givens itself',
        default='{}')
    ap.add_argument(
        '--results_path',
        help='Path to file where JSON with prediction results will be dumped',
        default='result_file.json')

    ap.add_argument(
        '--track_url',
        help='tracker url',
        default='http://tracker.mockup.url')

    ap.add_argument(
        '--debug_print',
        action='store_true',
        help='Printing results on screen')

    ap.add_argument(
        '--prettify_json',
        action='store_true',
        help='Pretty dump to JSON')

    ap.add_argument(
        '--full_output',
        action='store_true',
        help='dump to JSON all variants along with scores')

    ap.add_argument(
        '--mockup_track_endpoint',
        action='store_true',
        help='Should endpoint be mocked up')

    pa = ap.parse_args()

    if pa.mockup_track_endpoint:
        try:
            import requests_mock as rqm
        except ImportError as ierr:
            raise ImportError(
                'Please install `requests_mock` package to mock track endpoint:\npip3 install requests_mock')

    dm = DecisionModel(model_name=None, track_url=pa.track_url).load(model_url=pa.model_url)

    if pa.operation not in DecisionModel.SUPPORTED_CALLS:
        raise ValueError(
            'CLI supported method_calls are {}. \n Provided method_calls: {}'
            .format(','.join(dm.SUPPORTED_CALLS), pa.operation))

    if pa.operation != 'get':
        assert hasattr(dm, pa.operation)

    variants = json.loads(
        read_jsonstring_from_file(path_to_file=pa.variants) if os.path.isfile(pa.variants) else pa.variants)
    givens = json.loads(
        read_jsonstring_from_file(path_to_file=pa.givens) if os.path.isfile(pa.givens) else pa.givens)
    dc = DecisionContext(decision_model=dm, givens=givens)
    if pa.operation == 'get':

        scores = dc.score(variants=variants)
        decision = dc.decide(variants=variants, scores=scores)
        result = mockup_and_execute(
            executed_function=decision.get, function_kwargs={},
            mockup_track_endpoint=pa.mockup_track_endpoint, track_url=pa.track_url)
    else:
        scores = dm._score(variants=variants, givens=givens) \
            if pa.operation not in ["optimize", "choose_multivariate"] else None

        if pa.operation == 'score':
            result = scores
        else:
            desired_operation = getattr(dc, pa.operation)
            if pa.operation in ["rank", "which_from"]:
                kwargs = {'variants': variants}
            elif pa.operation in ["optimize", "choose_multivariate"]:
                kwargs = {'variant_map': variants}
            else:
                kwargs = {'variants': variants, 'scores': scores}

            # executed_function, function_kwargs, mockup_track_endpoint, track_url
            result = mockup_and_execute(
                executed_function=desired_operation, function_kwargs=kwargs,
                mockup_track_endpoint=pa.mockup_track_endpoint, track_url=pa.track_url)

        if isinstance(result, np.ndarray):
            result = result.tolist()

    result_payload = result
    if pa.full_output:
        if pa.operation == 'score':
            result_payload = [{'variant': k, 'score': v} for k, v in zip(variants, result)]
        elif pa.operation == 'rank':
            result_payload = [{'variant': k, 'score': v} for k, v in zip(result, list(reversed(sorted(scores))))]

    string_result = json.dumps(result_payload)

    if pa.prettify_json:
        try:
            import simplejson
        except ImportError as ierr:
            raise ImportError(
                'Please install `simplejson` package to mock track endpoint: pip3 install simplejson')
        string_result = simplejson.dumps(json.loads(string_result), indent=4)

    if pa.debug_print:
        print('\n##################################################################\n\n'
              'Result of method call: {} \non variants: {}\nwith model: {} \n\n'
              '##################################################################\n\n'
              'is: \n'
              '{}'
              .format(pa.operation, pa.variants, pa.model_url, result))

    if pa.results_path:
        with open(pa.results_path, 'w') as resf:
            resf.writelines(string_result)



from copy import deepcopy

import numpy as np

import improveai.decision as d
import improveai.decision_model as dm
from improveai.utils.general_purpose_tools import check_variants, get_variants_from_args


class DecisionContext:

    @property
    def decision_model(self):
        """
        DecisionModel of this DecisionContext

        Returns
        -------
        DecisionModel
            DecisionModel of this DecisionContext

        """
        return self._decision_model

    @decision_model.setter
    def decision_model(self, value):
        # DecisionContext's model must not be None and of desired type
        assert value is not None
        assert isinstance(value, dm.DecisionModel)
        self._decision_model = value

    @property
    def givens(self) -> dict:
        """
        Givens for this DecisionContext

        Returns
        -------
        dict
            Givens for this DecisionContext

        """
        return self._givens

    @givens.setter
    def givens(self, value):
        # givens can be None or dict
        assert isinstance(value, dict) or value is None
        self._givens = value

    def __init__(self, decision_model, givens: dict or None):
        """
        Init with params

        Parameters
        ----------
        decision_model: DecisionModel
            decision model for this DecisionContext
        givens: dict
            givens for this DecisionContext

        """
        self.decision_model = decision_model
        self.givens = givens

    def score(self, variants) -> np.ndarray:
        """
        Scores provided variants with available givens

        Parameters
        ----------
        variants: list or tuple or np.ndarray
            array of variants to choose from

        Returns
        -------
        np.ndarray
            array of float64 scores

        """

        # get givens from provider
        givens = self.decision_model.givens_provider.givens(for_model=self.decision_model, givens=self.givens)
        return self.decision_model._score(variants=variants, givens=givens)

    def which(self, *variants) -> tuple:
        """
        A short hand version of chooseFrom that returns the chosen result directly,
        automatically tracking the decision in the process. This would be a varargs
        version of chooseFrom that would also take an array as the only argument.
        If the only argument was not an array it would throw an exception.
        Since the Decision object is not returned, rewards would only be tracked
        through DecisionModel.addReward

        Parameters
        ----------
        *variants: list or tuple or np.ndarray
            array of variants passed as positional parameters

        Returns
        -------
        object, str
            a tuple of (<best variant>, <decision id>)

        """

        return self.which_from(variants=get_variants_from_args(variants))

    def which_from(self, variants: list or np.ndarray) -> tuple:
        """
        Makes a decision for provided variants and DecisionContext givens.

        Parameters
        ----------
        variants: list or np.ndarray
            variants from which best will be chosen

        Returns
        -------
        object, str
            a tuple of (<best variant>, <decision id>)

        """
        # create a Decision object
        decision = self.decide(variants=variants)
        # track if possible -> tracker and track_url exist
        if self.decision_model.track_url is not None and self.decision_model.tracker is not None:
            decision.track()
        # return best variant and decision ID to allow rewarding
        return decision.best, decision.id_

    def decide(self, variants: list or np.ndarray, scores: list or np.ndarray = None,
               ordered: bool = False):
        """
        Creates decision but does not track it by default.
        If not scores are provided and input variants are not ranked performs
        scoring and ranking. If scores are provided but variants are not ordered
        ranks variants before making a decision

        Parameters
        ----------
        variants: list or np.ndarray
            variants for the decision
        scores: list or np.ndarray
            scores for variants, nullable
        ordered: bool
            flag indicating if input variants are ordered

        Returns
        -------
        d.Decision
            decision object created for input data

        """
        # get givens via GivensProvider
        givens = self.decision_model.givens_provider.givens(for_model=self.decision_model, givens=self.givens)
        # rank variants if they are not ordered
        assert isinstance(ordered, bool)
        if scores is not None and ordered is True:
            raise ValueError('Both "scores" and "ordered" are not None. One of them must be None (please check docs).')

        if not ordered:
            # if variants are not ordered scoring or ranking must be performed
            if scores is None:
                # if scores were not provided they must be calculated
                scores = self.decision_model._score(variants=variants, givens=givens)

            # at this point scores must not be None and must be of the same length as variants
            assert scores is not None
            assert len(scores) == len(variants)
            # rank variants using scores
            ranked_variants = self.decision_model._rank(variants=variants, scores=scores)
        else:
            # variants are already ordered
            ranked_variants = deepcopy(variants)

        decision = d.Decision(
            decision_model=self.decision_model, ranked=ranked_variants, givens=givens)
        return decision

    def rank(self, variants: list or np.ndarray) -> list or np.ndarray:
        """
        Ranks provided variants from best to worst and creates decision from them

        Parameters
        ----------
        variants: list or np.ndarray
            variants to be ranked

        Returns
        -------
        list or np.ndarray
            ranked variants
        """
        return self.decide(variants=variants).ranked

    def optimize(self, variant_map: dict):
        """
        Get the best configuration for a given variants map

        Parameters
        ----------
        variant_map: dict
            mapping variants name -> variants list

        Returns
        -------
        object, str
            best variant and a decision ID

        """
        return self.which_from(
            variants=dm.DecisionModel.full_factorial_variants(variant_map=variant_map))

    def choose_first(self, variants: list or tuple or np.ndarray):
        # TODO method deprecated - will be removed in v8 upgrade
        """
        Chooses first from provided variants using gaussian scores (_generate_descending_gaussians())

        Parameters
        ----------
        variants: list or tuple or np.ndarray
            collection of variants passed as positional parameters

        Returns
        -------
        Decision
            A decision with first variants as the best one and gaussian scores
        """

        # make decision and do not track it
        return self.decide(variants=variants, ordered=True)

    def first(self, *variants) -> tuple:
        # TODO method deprecated - will be removed in v8 upgrade
        """
        Makes decision using first variant as best and tracks it.
        Accepts variants as pythonic *args

        Parameters
        ----------
        variants: list or tuple or np.ndarray
            collection of variants of which first will be chosen

        Returns
        -------
        object, str
            a tuple of (<first variant>, <decision id>)
        """
        # TODO think about get_variants_from_args() and checking variants for
        #  numpy types
        # make decision and track immediately
        decision = self.decide(variants=get_variants_from_args(variants), ordered=True)
        # return best and decision ID
        return decision.get(), decision.id_

    def choose_random(self, variants: list or tuple or np.ndarray):
        # TODO method deprecated - will be removed in v8 upgrade
        """
        Shuffles variants to return Decision with gaussian scores and random best variant

        Parameters
        ----------
        variants: list or tuple or np.ndarray
            collection of variants of which random will be chosen

        Returns
        -------
        Decision
            Decision with randomly chosen best variant
        """
        check_variants(variants)
        return self.decide(variants, scores=np.random.normal(size=len(variants)))

    def random(self, *variants) -> tuple:
        # TODO method deprecated - will be removed in v8 upgrade
        """
        Makes decision using randomly selected variant as best and tracks it.
        Accepts variants as pythonic *args

        Parameters
        ----------
        variants: list or np.ndarray
            collection of variants of which first will be chosen

        Returns
        -------
        object, str
            a tuple of (<random variant>, <decision id>)
        """
        # check variants before scoring
        check_variants(variants)
        # unpack variants -> avoid pitfall of [[variants]] in random scores generation
        unpacked_variants = get_variants_from_args(variants)
        # decide which one is best
        decision = self.decide(
            variants=unpacked_variants, scores=np.random.normal(size=len(unpacked_variants)))
        return decision.get(), decision.id_

    def choose_from(
            self, variants: list or tuple or np.ndarray, scores: np.ndarray or list or None):
        # TODO method deprecated - will be removed in v8 upgrade
        """
        Makes a Decision without tracking it

        Parameters
        ----------
        variants: list or tuple or np.ndarray
            array of variants to choose from
        scores: np.ndarray or list or None
            list of scores for variants

        Returns
        -------
        Decision
            snapshot of the Decision made with provided variants and available givens
        """
        return self.decide(variants=variants, scores=scores)

    def choose_multivariate(self, variant_map: dict) -> dict:
        """
        Chooses the best configuration from full factorial of input variants map

        Parameters
        ----------
        variant_map: dict
            map of key -> variants set

        Returns
        -------
        dict
            combination of the best variants in a dict

        """
        # check_variant_map(variant_map=variant_map)
        return self.choose_from(
            variants=dm.DecisionModel.full_factorial_variants(variant_map=variant_map),
            scores=None)

    def _track(self, variant: object, runners_up: list or np.ndarray, sample: object, sample_pool_size: int) -> str:
        """
        Tracks provided variant with runners up and sample

        Parameters
        ----------
        variant: object
            variant to be tracked
        runners_up: list
            runners_up to be tracked
        sample: object
            a sample for an input variant
        sample_pool_size: int
            number of variants from which sample was drawn

        Returns
        -------
        str
            decision ID
        """
        assert self.decision_model.track_url is not None
        assert self.decision_model.tracker is not None

        if runners_up is not None:
            check_variants(runners_up)

        if runners_up is None:
            runners_up_count = 0
        else:
            if not isinstance(runners_up, list):
                runners_up = list(runners_up)
            runners_up_count = len(runners_up)

        variants_count = 1 + runners_up_count + sample_pool_size
        body = self.decision_model.tracker._get_decision_track_body(
            variant=variant, model_name=self.decision_model.model_name, variants_count=variants_count,
            givens=self.givens, runners_up=runners_up, sample=sample,
            has_sample=True if sample_pool_size > 0 else False)

        return self.decision_model.tracker.post_improve_request(body_values=body)



import requests as rq
from urllib.parse import urlparse


def is_path_http_addr(pth_to_model: str) -> bool:
    """
    Checks if provided string is a http address

    Parameters
    ----------
    pth_to_model

    Returns
    -------

    """
    try:
        res = urlparse(pth_to_model)
        return all([res.scheme, res.netloc])
    except ValueError:
        return False


def get_model_bytes_from_url(model_url: str) -> bytes:
    """
    Gets model from provided URL

    Parameters
    ----------
    model_url: str
        url to get the model from

    Returns
    -------
    bytes
        downloaded bytes source url

    """

    try:
        headers = {"Accept-Encoding": "identity"}
        model_resp = rq.get(model_url, headers=headers)
        if model_resp.status_code != 200:
            raise ('Unable to load model from: {} path'.format(model_url))
    except Exception as exc:
        print('While getting model got: {} error'.format(exc))
        raise ValueError('Unable to load model from: {} path'.format(model_url))

    dled_model = model_resp.content

    if not isinstance(model_resp.content, bytes):
        raise TypeError('Downloaded model is not of a `bytes` type')

    return dled_model



from codecs import decode
from copy import deepcopy
import datetime
from ksuid import Ksuid
import numpy as np

ALLOWED_VARIANT_COLLECTION_TYPES = [list, tuple, np.ndarray]


def append_prefix_to_dict_keys(input_dict: dict, prfx: str) -> dict:
    """
    Appends prefix to the input dict keys

    Parameters
    ----------
    input_dict: dict
        dict which keys will be appended
    prfx: str
        prf to append to input dict keys

    Returns
    -------
    dict
        dict with keys appended

    """
    return dict(zip(
        ['{}{}'.format(prfx, key) for key in input_dict.keys()],
        list(input_dict.values())))


def impute_missing_dict_keys(
        all_des_keys: list, imputed_dict, imputer_value: float = np.nan) -> dict:
    """
    Imputes provided value into missing keys of input dict

    Parameters
    ----------
    all_des_keys: list
        list of all necessary keys in returned dict
    imputed_dict: dict
        input dict
    imputer_value: object
        value to be imputed for missing values

    Returns
    -------
    dict
        dict with all missing keys imputed

    """

    processed_encoded_features = deepcopy(imputed_dict)

    for feat_n in all_des_keys:
        if feat_n not in imputed_dict.keys():
            processed_encoded_features[feat_n] = imputer_value

    return processed_encoded_features


def read_jsonstring_from_file(
        path_to_file: str, mode: str = 'r', method: str = 'readlines',
        decode_escape_char: bool = False) -> str:
    """
    Safely reads desired file with json wrapper

    Parameters
    ----------
    path_to_file: str
        pth to loaded file
    mode: str
        read mode
    decode_escape_char: bool
        should double escape be repalced by single ?

    Returns
    -------
    str
        read json string

    """

    with open(path_to_file, mode) as rf:
        if method == 'readlines':
            contents = rf.readlines()
        elif method == 'read':
            contents = rf.read()
        else:
            raise ValueError('Wrong reading method value: {}'.format(method))

    if isinstance(contents, list):
        contents = ''.join(contents)

    if decode_escape_char:
        contents = decode(contents, 'unicode_escape')

    assert isinstance(contents, str)
    return contents


def sigmoid(x: float, logit_const: float) -> float:
    """
    Calculate sigmoid value

    Parameters
    ----------
    x: float
        arg passed to sigmoid function
    logit_const: float
        constant added to sigmoid argument

    Returns
    -------
    float
        value of sigmoid function

    """
    exp_arg = logit_const - x
    return 1 / (1 + np.exp(exp_arg))


def check_variants(variants: list or tuple or np.ndarray) -> list or tuple or np.ndarray:
    """
    Check if variants are of desired type and if they are not an empty collection

    Parameters
    ----------
    variants: list or tuple or np.ndarray
        checked variants

    """
    # following checks take only less than 0,001 ms
    assert variants is not None
    assert type(variants) in ALLOWED_VARIANT_COLLECTION_TYPES
    # raise if variants are an empty list
    assert len(variants) > 0, '`variants` must not be an empty collection'


def is_valid_variants_type(variants):
    return type(variants) in ALLOWED_VARIANT_COLLECTION_TYPES


def get_variants_from_args(variants: list or tuple or np.ndarray) -> list or tuple or np.ndarray:
    """
    Extract variants from pythonic args

    Parameters
    ----------
    variants: list or tuple or np.ndarray


    Returns
    -------

    """

    if len(variants) == 1:
        # if the first element is a list -> it is ok to use such list as variants
        # if the first element is not a list / tuple / np.ndarray raise
        assert isinstance(variants[0], list) or isinstance(variants[0], tuple) or isinstance(variants[0], np.ndarray)
        return variants[0]

    return [v if not is_object_numpy_type(v) else v.item() for v in variants]


def is_valid_ksuid(id_: str) -> bool:
    """
    Checks if input value is a valid Ksuid string

    Parameters
    ----------
    id_: str
        checked string

    Returns
    -------
    bool
        True if string is a valid ksuid otherwise False

    """
    if not isinstance(id_, str):
        return False

    if len(id_) != 27:
        return False

    try:
        # Disallow KSUIDs from the future, otherwise it could severely hurt
        # the performance of the partitions by creating a huge partition in the future
        # that new records keep aggregating into. At some point that partition would
        # no longer fit in RAM and processing could seize.
        if Ksuid.from_base62(id_).datetime > datetime.datetime.now(datetime.timezone.utc):
            return False
    except:
        # there was an exception parsing the KSUID, fail
        return False

    return True


def is_object_numpy_type(checked_object: object):
    """
    Check if a passed input object is a numpy type or not

    Parameters
    ----------
    checked_object: object
        object to be checked for numpy type

    Returns
    -------
    bool
        True if object is of a nupy type False otherwise

    """
    if type(checked_object).__module__ == np.__name__:
        return True
    return False



import numpy as np


def encoded_variant_to_np(
        encoded_variant: dict, feature_names: list or np.ndarray) -> np.ndarray:
    """
    Puts values from `encoded_variant` into numpy array. Only those features in `encoded_variant`
    which overlap with `feature_names`. Resulting array must contain all values stored in `feature_names`.
    Features missing in `encoded_variants` are filled with np.nan.

    Parameters
    ----------
    encoded_variant: dict
        fully encoded single variant
    feature_names: list or np.ndarray
        array of feature names from model which will be used for predictions

    Returns
    -------
    np.ndarray
        single row of a shape (1, num. features) which contains fully encoded variant

    """

    result = np.empty(shape=(len(feature_names), ))
    result[:] = np.nan
    hash_index_map = \
        {feature_hash: index for index, feature_hash
         in enumerate(feature_names)}

    filler = \
        np.array(
            [(hash_index_map.get(feature_name, None), value)
             for feature_name, value in encoded_variant.items()
             if hash_index_map.get(feature_name, None) is not None])
    if len(filler) > 0:
        result[filler[:, 0].astype(int)] = filler[:, 1]

    return result


def encoded_variants_to_np(
        encoded_variants: np.ndarray, feature_names: np.ndarray) -> np.ndarray:
    """
    Converts an array of variants into 2D numpy array (<N variants>, <M features>).
    Only those features in a single encoded variants which overlap with `feature_names`
    will be present in the resulting matrix. Resulting array must contain all values
    stored in `feature_names`. Features missing in `encoded_variants` are filled with np.nan.

    Parameters
    ----------
    encoded_variants: np.ndarray
        array of encoded variants
    feature_names: np.ndarray
        array of feature names from model which will be used for predictions

    Returns
    -------
    np.ndarray
        array of (num. variants, num. features) shape

    """

    if len(encoded_variants) == 0:
        return np.full(
            shape=(len(encoded_variants), len(feature_names)),
            fill_value=np.nan)

    encoded_variants_matrix = \
        np.empty((len(encoded_variants), len(feature_names)))

    encoded_variants_matrix[:] = [
        encoded_variant_to_np(
            encoded_variant=encoded_variant, feature_names=feature_names)
        for encoded_variant in encoded_variants]
    return encoded_variants_matrix



from copy import deepcopy
import numpy as np
import orjson
import re
import requests as rq
from typing import Dict
from warnings import warn

from ksuid import Ksuid

import improveai
from improveai.chooser import XGBChooser
from improveai.utils.general_purpose_tools import check_variants, is_valid_ksuid


class DecisionTracker:

    @property
    def MODEL_KEY(self) -> str:
        """
        Track request body key storing model name

        Returns
        -------
        str
            Track request body key storing model name

        """
        return "model"

    @property
    def MESSAGE_ID_KEY(self) -> str:
        """
        Track request body key storing message ID

        Returns
        -------
        str
            Track request body key storing message ID

        """
        return "message_id"

    @property
    def TYPE_KEY(self) -> str:
        """
        Track request body key storing request type (e.g. decision is a request type as well as reward)

        Returns
        -------
        str
            Track request body key storing request type

        """
        return "type"

    @property
    def VARIANT_KEY(self) -> str:
        """
        Track request body key storing best variant

        Returns
        -------
        str
            Track request body key storing best variant

        """
        return "variant"

    @property
    def GIVENS_KEY(self) -> str:
        """
        Track request body key storing givens

        Returns
        -------
        str
            Track request body key storing givens

        """
        return "givens"

    @property
    def VARIANTS_COUNT_KEY(self) -> str:
        """
        Track request body key storing variants count (from how many variants best was chosen)

        Returns
        -------
        str
            Track request body key storing variants count

        """
        return "count"

    @property
    def REWARD_TYPE(self) -> str:
        """
        If request is a reward this should be provided as `<request body>[<TYPE_KEY>]`

        Returns
        -------
        str
            type to be appended to request body in case of reward request

        """
        return 'reward'

    @property
    def REWARD_KEY(self) -> str:
        """
        Track request body key storing reward value

        Returns
        -------
        str
            Track request body key storing reward value

        """
        return 'reward'

    @property
    def DECISION_TYPE(self) -> str:
        """
        If a request is a decision this should be provided as `<request body>[<TYPE_KEY>]`

        Returns
        -------
        str
            type to be appended to request body in case of decision request

        """

        return "decision"

    @property
    def DECISION_ID_KEY(self) -> str:
        """
        Track request body key storing decision ID

        Returns
        -------
        str
            Track request body key storing decision ID

        """
        return "decision_id"

    @property
    def API_KEY_HEADER(self) -> str:
        """
        Track request headers key storing `API key`

        Returns
        -------
        str
            Track request headers key storing `API key`

        """

        return "x-api-key"

    @property
    def PAYLOAD_FOR_ERROR_KEY(self) -> str:
        """
        user info dict key storing track request body which caused an error

        Returns
        -------
        str
            user info dict key storing track request body which caused an error

        """
        return 'ERROR_WITH_PAYLOAD'

    @property
    def REQUEST_ERROR_CODE_KEY(self) -> str:
        """
        user info dict key storing track request error code

        Returns
        -------
        str
            user info dict key storing track request error code

        """
        return 'REQUEST_ERROR_CODE'

    @property
    def RUNNERS_UP_KEY(self) -> str:
        """
        Track request headers key storing runners up

        Returns
        -------
        str
            Track request headers key storing runners up

        """
        return 'runners_up'

    @property
    def VARIANTS_COUNT_KEY(self) -> str:
        """
        Track request headers key storing variants count

        Returns
        -------
        str
            Track request headers key storing variants count

        """
        return 'count'

    @property
    def SAMPLE_KEY(self) -> str:
        """
        Track request headers key storing sample

        Returns
        -------
        str
            Track request headers key storing sample

        """
        return 'sample'

    @property
    def track_url(self) -> str:
        """
        Improve AI track endpoint URL

        Returns
        -------
        str
            Improve AI track endpoint URL


        """
        return self._track_url

    @track_url.setter
    def track_url(self, new_val: str):
        assert new_val is not None
        self._track_url = new_val

    @property
    def api_key(self) -> str:
        """
        track endpoint API key (if applicable); Can be None

        Returns
        -------
        str
            track endpoint API key

        """
        return self._api_key

    @api_key.setter
    def api_key(self, new_val: str):
        self._api_key = new_val

    @property
    def max_runners_up(self):
        """
        maximum number of runners up to be included in the Improve AI request

        Returns
        -------
        int
            maximum number of runners up to be included in the Improve AI request

        """
        return self._max_runners_up

    @max_runners_up.setter
    def max_runners_up(self, new_val):
        self._max_runners_up = new_val

    def __init__(self, track_url: str, track_api_key: str = None):
        """
        Init with params

        Parameters
        ----------
        track_url: str
            Improve AI track endpoint URL
        track_api_key: str
            Improve AI track endpoint API key (nullable)
        """

        self.track_url = track_url
        self.api_key = track_api_key
        # defaults to 50
        self.max_runners_up = 50

    def _should_track_runners_up(self, variants_count: int) -> bool:
        """
        Returns bool that indicates whether runners up should be tracked


        Parameters
        ----------
        variants_count: int
            number of variants


        Returns
        -------
        bool
            flag indicating if runners up will be tracked

        """

        if variants_count == 1 or self.max_runners_up == 0:
            return False
        elif variants_count == 2:
            return True
        else:
            return np.random.rand() < 1 / min(variants_count - 1, self.max_runners_up)

    def _top_runners_up(
            self, ranked_variants: list or tuple or np.ndarray) -> list or tuple or np.ndarray or None:
        """
        Select top N runners up from `ranked_variants`


        Parameters
        ----------
        ranked_variants: list
            variants ordered descending by scores


        Returns
        -------
        Iterable or None
            None if there are no runners up to track otherwise list of tracked runners up

        """
        check_variants(ranked_variants)
        # len(ranked_variants) - 1 -> this will not include last element of collection
        top_runners_up = ranked_variants[1:min(len(ranked_variants), self.max_runners_up + 1)]\
            if ranked_variants is not None else None

        # if there is a positive max_runners_up and more than 1 variant there must be at least 1 runner up
        if len(ranked_variants) > 1 and self.max_runners_up > 0:
            assert top_runners_up is not None and len(top_runners_up) > 0

        # if max_runners_up == 0 there are no runners up
        if self.max_runners_up == 0:
            assert not top_runners_up  # None or []

        # If `top_runners_up` == [] return None
        if top_runners_up is None or len(top_runners_up) == 0:
            return None

        returned_top_runners_up = top_runners_up
        if isinstance(top_runners_up, np.ndarray):
            returned_top_runners_up = top_runners_up.tolist()

        return returned_top_runners_up

    def _is_sample_available(self, ranked_variants: list or None, runners_up: list or None) -> bool:
        """
        Returns True / False flag indicating whether sample is available


        Parameters
        ----------
        ranked_variants: list or None
            collection of evaluated variants
        runners_up: list or None
            tracked runners up


        Returns
        -------
        bool
            True if sample is available False otherwise

        """

        variants_count = len(ranked_variants)
        runners_up_count = len(runners_up) if runners_up else 0

        if variants_count - runners_up_count - 1 > 0:
            return True

        return False

    def _get_decision_track_body(
            self, variant: object, model_name: str, variants_count: int, givens: dict,
            runners_up: list, sample: object, has_sample: bool):
        """
        Helper method to create track body. used by DecisionTracker's track() and
        DecisionModel / DecisionContext track()

        Parameters
        ----------
        variant: object
            tracked variant
        model_name: str
            tracked model name
        variants_count: int
            number of variants
        givens: dict
            givens for this decision
        runners_up: list
            list of runners up to be tracked
        has_sample: bool
            has Decision a sample
        sample: object
            sample to be tracked

        Returns
        -------
        dict
            body for _post_improve_request()

        """
        assert re.search(XGBChooser.MODEL_NAME_REGEXP, model_name) is not None

        body = {
            self.TYPE_KEY: self.DECISION_TYPE,
            self.MODEL_KEY: model_name,
            self.VARIANT_KEY: variant,
            self.VARIANTS_COUNT_KEY: variants_count}

        if runners_up is not None:
            check_variants(runners_up)
            body[self.RUNNERS_UP_KEY] = runners_up

        if has_sample:
            body[self.SAMPLE_KEY] = sample

        if givens is not None:
            body[self.GIVENS_KEY] = givens

        return body

    def track(self, ranked_variants: list or np.ndarray, givens: dict, model_name: str) -> str or None:
        """
        Track that variant is causal in the system


        Parameters
        ----------
        ranked_variants: list or np.ndarray
            collection of tracked variants
        givens: dict:
            givens for variants
        model_name: str
            name of model which made the decision (?) / to which observations
            will be assigned

        Returns
        -------
        str or None
            message id of sent improve request or None if an error happened

        """

        assert self.track_url is not None

        if model_name is None:
            warn('`model_name` must not be None in order to be tracked')
            return None

        if not re.search(XGBChooser.MODEL_NAME_REGEXP, model_name):
            warn(f'`model_name` must comply with {XGBChooser.MODEL_NAME_REGEXP} regexp')
            return None

        if isinstance(ranked_variants, np.ndarray):
            ranked_variants = ranked_variants.tolist()

        variants_count = len(ranked_variants) if ranked_variants is not None else 1
        track_runners_up = self._should_track_runners_up(variants_count=variants_count)

        if len(ranked_variants) == 2 and self.max_runners_up > 0:
            # for 2 variants runners_up should be True
            assert track_runners_up

        # for track_runners_up == false (max_runners_up == 0)
        # we skip this clause and runners_up == None
        runners_up = None
        if track_runners_up:
            runners_up = self._top_runners_up(ranked_variants=ranked_variants)

        # If runners_up == None and len(variants) == 2 -> sample should be extracted
        sample = None
        has_sample = self._is_sample_available(ranked_variants=ranked_variants, runners_up=runners_up)
        if has_sample:
            sample = self.get_sample(ranked_variants=ranked_variants, track_runners_up=track_runners_up)

        body = self._get_decision_track_body(
            variant=ranked_variants[0], model_name=model_name, variants_count=variants_count,
            givens=givens, runners_up=runners_up, sample=sample, has_sample=has_sample)

        return self.post_improve_request(body_values=body)

    def add_reward(self, reward: float or int, model_name: str, decision_id: str):
        """
        Adds provided reward for a given decision id made by a given model.


        Parameters
        ----------
        reward: float or int
            reward to be assigned to a given decision
        model_name: str
            name of a model which made rewarded decision
        decision_id: str
            ksuid of rewarded decision


        Returns
        -------
        str
            message ID

        """

        assert self.track_url is not None, '`track_url` is None - please provide valid `track_url`'
        assert model_name is not None and decision_id is not None
        assert is_valid_ksuid(decision_id)
        assert re.search(XGBChooser.MODEL_NAME_REGEXP, model_name) is not None
        assert isinstance(reward, float) or isinstance(reward, int)
        assert reward is not None
        assert not np.isnan(reward)
        assert not np.isinf(reward)

        body = {
            self.TYPE_KEY: self.REWARD_TYPE,
            self.MODEL_KEY: model_name,
            self.REWARD_KEY: reward,
            self.DECISION_ID_KEY: decision_id}

        self.post_improve_request(body_values=body)

    def get_sample(self, ranked_variants: list, track_runners_up: bool) -> object:
        """
        Gets sample from ranked_variants. Takes runenrs up into account


        Parameters
        ----------
        ranked_variants: list or np.ndarray
            list of ranked variants
        track_runners_up: bool
            should runners up be tracked ?


        Returns
        -------
        object
            sample

        """

        check_variants(ranked_variants)
        assert isinstance(track_runners_up, bool)
        assert len(ranked_variants) > 1

        if len(ranked_variants) == 2:
            assert self.max_runners_up == 0

        # If there are no runners up, then sample is a random sample
        # from variants with just best excluded.

        # randomly select variants from all minus best
        if not track_runners_up:
            return ranked_variants[np.random.randint(1, len(ranked_variants))]

        # If there are no remaining variants after best and runners up,
        # then there is no sample.
        # below is just a check
        last_runner_up_idx = min(len(ranked_variants), self.max_runners_up + 1)
        assert last_runner_up_idx < len(ranked_variants)

        # If there are runners up, then sample is a random sample from
        # variants with best and runners up excluded.
        return ranked_variants[np.random.randint(last_runner_up_idx, len(ranked_variants))]

    def _is_valid_message_id(self, message_id: str or None) -> bool:
        """
        Check if message_id is  valid


        Parameters
        ----------
        message_id: str
            checked message_id


        Returns
        -------
        bool
            True if message_id otherwise False

        """
        try:
            assert isinstance(message_id, str) or message_id is None
            Ksuid.from_base62(message_id)
            return True
        except:
            return False

    def do_post_improve_request(self, payload_json: str, headers: dict):
        """
        Execute improve POST request with provided payload and headers

        Parameters
        ----------
        payload_json: str
            improveai body dumped to JSON
        headers: dict
            request headers

        Returns
        -------
        None
            None

        """
        try:
            response = rq.post(url=self.track_url, data=payload_json, headers=headers)
            if response.status_code >= 400:
                user_info = dict(deepcopy(response.headers))
                user_info[self.REQUEST_ERROR_CODE_KEY] = str(response.status_code)

                if payload_json:
                    user_info[self.PAYLOAD_FOR_ERROR_KEY] = payload_json \
                        if len(payload_json) <= 1000 else payload_json[:100]

                # TODO test this path within thread but how?
                warn(
                    'When attempting to post to improve.ai endpoint got an error with code {} and user info: {}'
                    .format(str(response.status_code), orjson.dumps(user_info).decode('utf-8')))

        except Exception as exc:
            print('Following error occurred:')
            print(exc)

    def post_improve_request(self, body_values: Dict[str, object], message_id: str = None) -> str:
        """
        Posts request to tracker endpoint
        Parameters
        ----------
        body_values: dict
            dict to be posted
        block: callable
            callable to execute on error
        message_id: str or None
            ksuid of a given request
        Returns
        -------
        str
            message id of sent improve request
        """

        headers = {'Content-Type': 'application/json'}
        if self.api_key:
            headers[self.API_KEY_HEADER] = self.api_key

        assert self._is_valid_message_id(message_id=message_id)
        body = {self.MESSAGE_ID_KEY: message_id if message_id is not None else str(Ksuid())}
        body.update(body_values)

        # serialization is a must-have for this requests
        try:
            payload_json = orjson.dumps(body).decode('utf-8')
        except Exception as exc:
            warn("Data serialization error: {}\nbody: {}".format(exc, body))
            return None

        improveai.track_improve_executor.submit(self.do_post_improve_request, *[payload_json, headers])

        return body[self.MESSAGE_ID_KEY]



from concurrent.futures import ThreadPoolExecutor
from pkg_resources import get_distribution, DistributionNotFound
from warnings import warn

from improveai.decision import Decision
from improveai.decision_context import DecisionContext
from improveai.feature_encoder import FeatureEncoder
from improveai.decision_model import DecisionModel, load_model
from improveai.givens_provider import GivensProvider
# import gcc and python3-dev detection utility
from improveai.settings import CYTHON_BACKEND_AVAILABLE, MAX_TRACK_THREADS

if CYTHON_BACKEND_AVAILABLE:
    # if gcc and python3 dev installed use cythonized FeatureEcoder
    from improveai.cythonized_feature_encoding import cfe
    FeatureEncoder = cfe.FeatureEncoder
else:
    # no cythonized FeatureEncoder
    warn('No build tools detected (gcc, python3-dev) -> FastFeatureEncoder falls back to normal FeatureEncoder')

# set __version__ package attribute; If package is not installed with pip set INFO string to __version__
try:
    __version__ = get_distribution('improveai').version
except DistributionNotFound as enferr:
    __version__ = 'Distribution not found - please install `improveai` wit pip'

track_improve_executor = ThreadPoolExecutor(max_workers=MAX_TRACK_THREADS)



import os


def gcc_and_py3_dev_installed() -> bool:
    """
    Checks if gcc and python3 headers are installed.

    Returns
    -------
    bool
        True if both gcc and python3 headers are installed otherwise False

    """
    gcc_installed = os.system('gcc -v > /dev/null 2>&1') == 0
    py3_dev_installed = os.system('python3-config --help > /dev/null 2>&1') == 0
    return gcc_installed and py3_dev_installed


# docs for module level variables like CYTHON_BACKEND_AVAILABLE are placed below them
CYTHON_BACKEND_AVAILABLE = gcc_and_py3_dev_installed()
"""
Indicates if cython backend is usable with python-SDK
"""


DEBUG = False
"""
Indicates if debug level messages should be shown
"""

IMPROVE_ABS_PATH = os.sep.join(__file__.split(os.sep)[:-1])
MAX_TRACK_THREADS = 16



#!python3
#cython: language_level=3

cdef extern from "npy_no_deprecated_api.h": pass

import cython
import numpy as np
cimport numpy as np


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef encoded_variant_into_np_row(
        dict encoded_variant, list feature_names, np.ndarray into):
    """
    Fills in `into` row using provided variants and `feature_names`

    Parameters
    ----------
    encoded_variant: dict
        dict containing fully encoded variant (variant + givens)
    feature_names: np.ndarray or list
        collection of desired feature names
    into: np.ndarray
        array to be filled with encoded

    Returns
    -------
    None
        None

    """

    cdef dict hash_index_map = \
        {feature_hash: index for index, feature_hash
         in enumerate(feature_names)}

    cdef np.ndarray filler = \
        np.array(
            [(hash_index_map.get(feature_name, None), value)
             for feature_name, value in encoded_variant.items()
             if hash_index_map.get(feature_name, None) is not None])

    # sum into with encoded variants treating nans in sums as zeros
    cdef np.ndarray subset_index = np.empty(len(filler))

    if len(filler) > 0:

        subset_index = filler[:, 0].astype(int)

        into[subset_index] = np.nansum(
            np.array([into[subset_index], filler[:, 1]]), axis=0)


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef np.ndarray encode_variants_single_givens(
        list variants, dict givens, double noise,
        object variants_encoder, object givens_encoder):
    """
    Encodes provided variants to a 2D array with single givens
    
    Parameters
    ----------
    variants: list
        list of encoded variants
    givens: dict
        givens to be encoded with each variant
    noise: float
        value within 0 - 1 range which will be 'shrunk' and added to feature value
    variants_encoder: callable
        a callable to be used as a variants encoder
    givens_encoder: callable
        a callable to be used as a givens encoder 
    
    Returns
    -------
    np.ndarray
        2D array with encoded variants; missing values are filled with np.nan

    """

    cdef dict into = {}

    cdef dict encoded_givens = \
        givens_encoder(givens, noise, into=dict(into))

    cdef int variants_count = len(variants)
    cdef np.ndarray res = np.empty((variants_count, ), dtype=object)

    cdef dict fully_encoded_variant = {}

    for variant_idx in range(variants_count):

        fully_encoded_variant = \
            variants_encoder(
                variants[variant_idx], noise, into=dict(encoded_givens))

        res[variant_idx] = fully_encoded_variant

    return res

@cython.boundscheck(False)
@cython.wraparound(False)
cpdef np.ndarray encode_variants_multiple_givens(
        list variants, list multiple_givens, list multiple_extra_features,
        double noise, object variants_encoder, object givens_encoder):
    """
    Encode each variant with corresponding givens
    
    Parameters
    ----------
    variants: list
        list of variants to be encoded
    multiple_givens: list
        list of givens - each element corresponds to a single variant which 
    multiple_extra_features: list
        list of extra features  - each element is a set of extra features for a single variant - givens pair
    noise: float
        value within 0 - 1 range which will be 'shrunk' and added to feature value
    variants_encoder: callable
        a callable to be used as a variants encoder
    givens_encoder: callable
        a callable to be used as a givens encoder

    Returns
    -------
    np.ndarray
        2D array with encoded variants; missing values are filled with np.nan

    """

    assert len(variants) == len(multiple_givens)
    cdef int records_count = len(variants)
    cdef np.ndarray res = np.empty((records_count,), dtype=object)

    cdef dict empty_into = {}

    cdef dict givens = {}
    cdef dict extra_features = {}
    cdef dict encoded_givens = {}
    cdef dict fully_encoded_variant = {}

    for variant_idx in range(records_count):

        encoded_givens = {}
        givens = multiple_givens[variant_idx]
        if givens:
            encoded_givens = \
                givens_encoder(givens, noise, into=dict(empty_into))

        res[variant_idx] = \
            variants_encoder(variants[variant_idx], noise, into=encoded_givens)

        extra_features = multiple_extra_features[variant_idx]
        if extra_features:
            res[variant_idx].update(extra_features)

    return res


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef double[:, :] encoded_variants_to_np(
        np.ndarray encoded_variants, list feature_names):
    """
    For each of `encoded_variants` selects features overlapping with `feature_names`, 
    fill missing features with np.nan and returns 2D array with all variants      
    
    Parameters
    ----------
    encoded_variants: np.ndarray
        as array with all variants to be encoded
    feature_names: list
        list of desires feature names

    Returns
    -------
    np.ndarray
        a 2D array with all encoded variants

    """

    cdef np.ndarray encoded_variants_array = \
        np.empty(shape=(len(encoded_variants), len(feature_names)), dtype=float)

    cdef dict name_to_index = \
        {feature_name: feature_index for feature_index, feature_name
         in enumerate(feature_names)}

    encoded_variants_array[:, :] = np.nan

    for encoded_variant_ixd, processed_variant in enumerate(encoded_variants):

        # this assumes that we have much less features per single variant that
        # total number of features
        filler = \
            np.array(
                [(name_to_index.get(feature_name, None), value)
                 for feature_name, value in processed_variant.items()
                 if name_to_index.get(feature_name, None) is not None])

        if len(filler) > 0:
            encoded_variants_array[encoded_variant_ixd, filler[:, 0]\
                .astype(int)] = filler[:, 1]

    return encoded_variants_array



import os
from setuptools import find_packages, setup


def gcc_and_py3_dev_installed():
    gcc_installed = os.system('gcc -v > /dev/null 2>&1') == 0
    py3_dev_installed = os.system('python3-config --help > /dev/null 2>&1') == 0
    return gcc_installed and py3_dev_installed


CYTHON_MODULE_DIR = 'cythonized_feature_encoding'
EXTENSION = 'pyx'
IMPROVE_DIR = 'improveai'


if __name__ == '__main__':

    install_requires = [
        "Cython>=0.29.14",
        "xxhash",
        "requests>=2.24.0",
        "numpy",
        "xgboost==1.4.2",
        "orjson",
        "svix-ksuid"]

    setup_kwargs = {
        'packages': find_packages(exclude=['*.tox*', '*tests*']),
        'install_requires': install_requires,
        'include_package_data': True}

    if gcc_and_py3_dev_installed():
        from Cython.Build import cythonize
        import numpy as np
        from setuptools import Extension

        cython_feature_encoding_utils_path_str = \
            os.sep.join(
                [IMPROVE_DIR, CYTHON_MODULE_DIR, 'cythonized_feature_encoding_utils.{}'.format(EXTENSION)])

        cython_feature_encoding_utils_ext = \
            Extension(
                '{}.{}.cythonized_feature_encoding_utils'.format(IMPROVE_DIR, CYTHON_MODULE_DIR),
                sources=[cython_feature_encoding_utils_path_str],
                include_dirs=[np.get_include(), os.sep.join(['.', IMPROVE_DIR, CYTHON_MODULE_DIR])])

        cython_feature_encoder_path_str = \
            os.sep.join(
                [IMPROVE_DIR, CYTHON_MODULE_DIR, 'cythonized_feature_encoder.{}'.format(EXTENSION)])

        cython_feature_encoder_ext = \
            Extension(
                '{}.{}.cythonized_feature_encoder'.format(IMPROVE_DIR, CYTHON_MODULE_DIR),
                sources=[cython_feature_encoder_path_str],
                include_dirs=[np.get_include(), os.sep.join(['.', IMPROVE_DIR, CYTHON_MODULE_DIR])])

        # append setup kwargs with cythonization info
        setup_kwargs.update(
            {'ext_modules': cythonize([cython_feature_encoding_utils_ext, cython_feature_encoder_ext], language_level="3"),
             'include_dirs': [np.get_include(), '.']
             })

    setup(**setup_kwargs)



from collections.abc import Iterable
import math
import numpy as np
import warnings
import xxhash


xxhash3 = xxhash.xxh3_64_intdigest
JSON_SERIALIZABLE_TYPES = {int, float, str, bool, list, tuple, dict, type(None)}
WARNED_ABOUT_ARRAY_ENCODING = False


class FeatureEncoder:
    """
    This class serves as a preprocessor which allows to
    convert event data in form of a JSON line (variants, givens and
    extra features) to the tabular format expected on input by XGBoost.

    Short summary of methods:
     - encode_givens() -> encodes givens / variant metadata to dict of
       feature name -> float pairs

     - encode_variant() -> encodes variant to dict of feature name -> float pairs

     - encode_feature_vector() -> encodes provided variant and givens;
       appends extra features to the encoded variant dict;
       converts encoded variant to numpy array using provided feature names

     - add_noise() -> sprinkles each value of the input dict with small noise

     - encode_to_np_matrix()  -> encodes collection of variants with collection
       of givens and extra features to the numpy 2D array / matrix. If desired
       uses cython backed (works much faster)

     - add_extra_features() - helper method for adding extra features to already
       encoded variants

    """

    VARIANT_METADATA_PARAMETER_NAME = 'givens'

    def __init__(self, model_seed):
        """
        Init with params

        Parameters
        ----------
        model_seed: int
            model seed to be used during feature encoding
        """

        if (model_seed < 0):
            raise TypeError(
                "xxhash3_64 takes an unsigned 64-bit integer as seed. "
                "Seed must be greater than or equal to 0.")

        # memoize commonly used seeds
        self.variant_seed = xxhash3("variant", seed=model_seed)
        self.value_seed = xxhash3("$value", seed=self.variant_seed)
        self.givens_seed = xxhash3("givens", seed=model_seed)
        self.__user_warned_about_array_encoding = False

    def _check_noise_value(self, noise: float):
        """
        Check whether noise value is valid; Raises if `noise` is invalid

        Parameters
        ----------
        noise: float
            checked value of no ise

        Returns
        -------
        None
            None
        """

        if noise > 1.0 or noise < 0.0:
            raise ValueError(
                'Provided `noise` is out of allowed bounds <0.0, 1.0>')

    def encode_givens(
            self, givens: dict or None, noise: float = 0.0, into: dict = None) -> dict:
        """
        Encodes provided givens of arbitrary complexity to a flat feature dict;
        Givens must be JSON encodable

        Parameters
        ----------
        givens: dict or None
            givens to be encoded
        noise: float
            value within 0 - 1 range which will be 'shrunk' and added to feature value
        into: dict
            a dict into which new features will be inserted; Can be None

        Returns
        -------
        dict
            A dict with results of givens encoding

        """

        self._check_noise_value(noise=noise)

        if givens is not None and not isinstance(givens, dict):
            raise TypeError(
                "Only dict type is supported for context encoding. {} type was "
                "provided.".format(type(givens)))

        if into is None:
            into = {}

        encode(givens, self.givens_seed, shrink(noise), into)
        return into

    def encode_variant(
            self, variant: object, noise: float = 0.0, into: dict = None) -> dict:
        """
        Encodes provided variant of arbitrary complexity to a flat feature dict;
        Variant must be JSON encodable

        Parameters
        ----------
        variant: object
            variant to be encoded
        noise: float
            value within 0 - 1 range which will be 'shrunk' and added to feature value
        into: dict
            a dict into which new features will be inserted; Can be None

        Returns
        -------
        dict
            A dict with results of variant encoding

        """

        self._check_noise_value(noise=noise)

        if into is None:
            into = {}

        small_noise = shrink(noise)

        if isinstance(variant, dict):
            encode(variant, self.variant_seed, small_noise, into)
        else:
            encode(variant, self.value_seed, small_noise, into)

        return into

    def encode_feature_vector(
            self, variant: object = None, givens: dict = None,
            extra_features: dict = None, feature_names: list or np.ndarray = None,
            noise: float = 0.0, into: np.ndarray = None):
        """
        Fully encodes provided variant and givens into a np.ndarray provided as `into` parameter.
        `into` must not be None

        Parameters
        ----------
        variant: object
            a JSON encodable object to be encoded to flat features' dict
        givens: dict
            a dict with givens to be enncoded to flat features' dict (all entries must be JSON encodable)
        extra_features: dict
            features to be added to encoded variant with givens
        feature_names: list or np.ndarray
            list of model's feature names (only overlapping features will be selected for resulting vector)
        noise: float
            value within 0 - 1 range which will be 'shrunk' and added to feature value
        into: np.ndarray
            an array into which feature values will be added / inserted

        Returns
        -------
        None
            None

        """

        if into is None:
            raise ValueError('`into` can`t be None')

        # encode givens
        encoded_givens = \
            self.encode_givens(givens=givens, noise=noise, into=dict())
        # encoded variant and givens
        encoded_variant = \
            self.encode_variant(
                variant=variant, noise=noise, into=encoded_givens)

        if extra_features:
            if not isinstance(extra_features, dict):
                raise TypeError(
                    'Provided `extra_features` is not a dict but: {}'
                    .format(extra_features))

            encoded_variant.update(extra_features)

        # n + nan = nan so you'll have to check for nan values on into
        hash_index_map = \
            {feature_hash: index for index, feature_hash in enumerate(feature_names)}

        filler = \
            np.array(
                [(hash_index_map.get(feature_name, None), value)
                 for feature_name, value in encoded_variant.items()
                 if hash_index_map.get(feature_name, None) is not None])

        # sum into with encoded variants treating nans in sums as zeros
        if len(filler) > 0:
            subset_index = filler[:, 0].astype(int)

            into[subset_index] = np.nansum(
                np.array([into[subset_index], filler[:, 1]]), axis=0)

    def encode_to_np_matrix(
            self, variants: Iterable, multiple_givens: Iterable,
            multiple_extra_features: Iterable, feature_names: Iterable,
            noise: float) -> np.ndarray:
        """
        Provided list of variants and corresponding lists of givens and extra
        features encodes variants completely and converts to numpy array

        Parameters
        ----------
        variants: list
            list of variants to be encoded
        multiple_givens: list
            list of givens - givens[i] will be used to encode variants[i]
        multiple_extra_features: list
            list of extra_features - multiple_extra_features[i] will be used
            to encode variants[i]
        feature_names: list
            names of features expected by the model
        noise: float
            noise to be used when encoding data

        Returns
        -------
        np.ndarray
            2D numpy array of encoded variants

        """

        if not isinstance(variants, list):
            variants = list(variants)

        if not isinstance(multiple_givens, list):
            multiple_givens = list(multiple_givens)

        if not isinstance(multiple_extra_features, list):
            multiple_extra_features = list(multiple_extra_features)

        if not isinstance(feature_names, list):
            feature_names = list(feature_names)

        encoded_variants_array = np.full((len(variants), len(feature_names)), np.nan)

        [self.encode_feature_vector(
            variant=variant, givens=givens, extra_features=extra_features,
            feature_names=feature_names, noise=noise, into=into)
         for variant, givens, extra_features, into
         in zip(variants, multiple_givens, multiple_extra_features,
                encoded_variants_array)]
        return encoded_variants_array

    def add_extra_features(self, encoded_variants: list, extra_features: list):
        """
        Once variants are encoded this method can be used to quickly append
        extra features

        Parameters
        ----------
        encoded_variants: list
            collection of encoded variants to be updated with extra features
        extra_features: list
            payload to be appended to encoded variants

        Returns
        -------
        None
            None

        """

        if extra_features is None:
            return

        if not isinstance(encoded_variants, list):
            raise TypeError('`encoded_variants` should be of a list type')

        if not isinstance(extra_features, list):
            raise TypeError('`extra_features` should be of a list type')

        [encoded_variant.update(single_extra_features)
         if single_extra_features is not None else encoded_variant
         for encoded_variant, single_extra_features in
         zip(encoded_variants, extra_features)]


def _is_object_json_serializable(object_) -> bool:
    """
    Checks if input value is JSON serializable

    Parameters
    ----------
    object_: object
        object to be checked

    Returns
    -------
    bool
        True if input is JSON serializable False otherwise
    """
    # try - except was slower
    return type(object_) in JSON_SERIALIZABLE_TYPES


def _has_top_level_string_keys(checked_dict) -> bool:
    """
    Check if all top level keys are of a string type. This is a helper function
    for encode() and since encode recurs in case of nested dicts only top level
    keys need to be checked. Return False on first encountered non-string key

    Parameters
    ----------
    checked_dict: dict
        dict which top level keys will be checked

    Returns
    -------
    bool
        True if all keys are strings otherwise False

    """

    return all(isinstance(k, str) for k in checked_dict.keys())


def warn_about_array_encoding(object_):
    """
    If object is an array warning will be printed (only once so the encoding speed does not suffer)

    Parameters
    ----------
    object_: object
        encoded object

    Returns
    -------
    None
        None
    """
    global WARNED_ABOUT_ARRAY_ENCODING

    if not WARNED_ABOUT_ARRAY_ENCODING and \
            (isinstance(object_, list) or isinstance(object_, tuple) or isinstance(object_, np.ndarray)):
        warnings.warn('Array encoding may change in the near future')
        WARNED_ABOUT_ARRAY_ENCODING = True


def encode(object_, seed, small_noise, features):
    """
    Encodes a JSON serializable object to  a flat key - value pair(s) structure / dict
    (sometimes a single `object_` will result in 2 features, e.g. strings, lists, etc...).
    Rules of encoding go as follows:

    - None, json null, {}, [], nan, are treated as missing feature_names and ignored. NaN is technically not allowed in JSON but check anyway.

    - boolean true and false are encoded as 1.0 and 0.0 respectively.

    - strings are encoded into an additional one-hot feature.

    - numbers are encoded as-is with up to about 5 decimal places of precision

    - a small amount of noise is incorporated to avoid overfitting

    - feature names are 8 hexadecimal characters


    Parameters
    ----------
    object_: object
        a JSON serializable object to be encoded to a flat key-value structure
    seed: int
        seed for xxhash3 to generate feature name
    small_noise: float
        a shrunk noise to be added to value of encoded feature
    features: dict
        a flat dict of {<feature name>: <feature value>, ...} pairs

    Returns
    -------
    None
        None

    """

    assert _is_object_json_serializable(object_)
    warn_about_array_encoding(object_=object_)

    if isinstance(object_, (int, float)):  # bool is an instanceof int
        if math.isnan(object_):  # nan is treated as missing feature, return
            return

        feature_name = hash_to_feature_name(seed)

        previous_object_ = \
            _get_previous_value(feature_name=feature_name, into=features, small_noise=small_noise)

        features[feature_name] = sprinkle(object_ + previous_object_, small_noise)

        return

    if isinstance(object_, str):
        hashed = xxhash3(object_, seed=seed)

        feature_name = hash_to_feature_name(seed)

        previous_hashed = \
            _get_previous_value(feature_name=feature_name, into=features, small_noise=small_noise)

        features[feature_name] = \
            sprinkle((((hashed & 0xffff0000) >> 16) - 0x8000) + previous_hashed, small_noise)

        hashed_feature_name = hash_to_feature_name(hashed)

        previous_hashed_for_feature_name = \
            _get_previous_value(feature_name=hashed_feature_name, into=features, small_noise=small_noise)

        features[hashed_feature_name] = \
            sprinkle(((hashed & 0xffff) - 0x8000) + previous_hashed_for_feature_name, small_noise)

        return

    if isinstance(object_, dict):
        assert _has_top_level_string_keys(object_)
        for key, value in object_.items():
            encode(value, xxhash3(key, seed=seed), small_noise, features)
        return

    if isinstance(object_, (list, tuple)):
        for index, item in enumerate(object_):
            encode(item, xxhash3(index.to_bytes(8, byteorder='big'), seed=seed), small_noise, features)
        return
    # None, json null, or unsupported type. Treat as missing feature, return


def _get_previous_value(
        feature_name: str, into: dict, small_noise: float):
    """
    Gets previous, 'unsprinkled' value of <feature_name> feature
    from <features> dict
    
    Parameters
    ----------
    feature_name: str
        name of a feature which previous value is desired
    into: dict
        dict containing current results of encoding
    small_noise: float
        small noise used to sprinkle

    Returns
    -------
    float
        'unsprinkled' value of desired feature

    """

    if feature_name not in into:
        return 0.0
    else:
        return reverse_sprinkle(into[feature_name], small_noise)


def hash_to_feature_name(hash_: int):
    """
    Converts a hash to string which will become a feature name

    Parameters
    ----------
    hash_: int
        an integer output from xxhash3

    Returns
    -------
    str
        a string representation of a hex feature name created from int

    """
    return '%0*x' % (8, (hash_ >> 32))


def shrink(noise):
    """
    Shrinks noise value by a hardcoded factor 2 ** -17

    Parameters
    ----------
    noise: float
        value within 0 - 1 range which will be 'shrunk' and added to feature value


    Returns
    -------
    float
        a shrunk noise

    """
    return noise * 2 ** -17


def sprinkle(x, small_noise):
    """
    Slightly modified input valuse using `small_noise`: (x + small_noise) * (1 + small_noise)

    Parameters
    ----------
    x: float
        a number to be 'modified'
    small_noise: float
        a small number with which `x` will be modified

    Returns
    -------
    float
        x modified with `small_noise`

    """
    return (x + small_noise) * (1 + small_noise)


def reverse_sprinkle(sprinkled_x, small_noise):
    """
    Retrieves true value from sprinkled one

    Parameters
    ----------
    sprinkled_x: float
        sprinkled value
    small_noise: float
        noise used to calculate sprinkled_Xx

    Returns
    -------
    float
        true value of x which was used to calculate sprinkled_x

    """
    return sprinkled_x / (1 + small_noise) - small_noise
